# coding=utf-8
# Copyright 2013-2016, 2019 Intel Corporation
# Copyright 2013, 2014 Advanced Micro Devices
# Copyright 2014 VMWare

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Generate text summaries to be printed to the console."""

import textwrap
import time

from framework import grouptools, backends
from .common import Results

__all__ = [
    'console',
]

_SUMMARY_TEMPLATE = textwrap.dedent("""\
    summary:
           name: {names}
           ----  {divider}
           pass: {pass_}
           fail: {fail}
          crash: {crash}
           skip: {skip}
        timeout: {timeout}
           warn: {warn}
     incomplete: {incomplete}
     dmesg-warn: {dmesg_warn}
     dmesg-fail: {dmesg_fail}
        changes: {changes}
          fixes: {fixes}
    regressions: {regressions}
          total: {total}
           time: {time}""")


def _print_summary(results):
    """print a summary."""

    lens = [max(min(len(x.name), 20), 8) for x in results.results]
    print_template = ' '.join(
        (lambda x: '{: >' + '{0}.{0}'.format(x) + '}')(y) for y in lens)

    def status_printer(stat):
        totals = [str(x.totals['root'][stat]) for x in results.results]
        return print_template.format(*totals)

    def elapsed_time(time_elapsed):
        """Return elapsed time (in seconds) as a string in HH:MM:SS format."""
        diff = time_elapsed.end - time_elapsed.start
        return time.strftime("%H:%M:%S", time.gmtime(diff))

    print(_SUMMARY_TEMPLATE.format(
        names=print_template.format(*[r.name for r in results.results]),
        divider=print_template.format(*['-'*l for l in lens]),
        pass_=status_printer('pass'),
        crash=status_printer('crash'),
        fail=status_printer('fail'),
        skip=status_printer('skip'),
        timeout=status_printer('timeout'),
        warn=status_printer('warn'),
        incomplete=status_printer('incomplete'),
        dmesg_warn=status_printer('dmesg-warn'),
        dmesg_fail=status_printer('dmesg-fail'),
        changes=print_template.format(
            *[str(s) for s in results.counts.changes]),
        fixes=print_template.format(
            *[str(s) for s in results.counts.fixes]),
        regressions=print_template.format(
            *[str(s) for s in results.counts.regressions]),
        total=print_template.format(*[
            str(sum(x.totals['root'].values()))
            for x in results.results]),
        time=print_template.format(
            *[elapsed_time(r.time_elapsed) for r in results.results]),
    ))

def _print_result(results, list_):
    """Takes a list of test names to print and prints the name and result."""
    for test in sorted(list_):
        print("{test}: {statuses}".format(
            test=grouptools.format(test),
            statuses=' '.join(str(r) for r in results.get_result(test))))


def console(resultsFiles, mode):
    """ Write summary information to the console for the given list of
    results files in the given mode."""
    assert mode in ['summary', 'diff', 'incomplete', 'fixes', 'problems', 'regressions', 'all'], mode
    results = Results([backends.load(r) for r in resultsFiles])

    # Print the name of the test and the status from each test run
    if mode == 'all':
        _print_result(results, results.names.all)
        _print_summary(results)
    elif mode == 'diff':
        _print_result(results, results.names.all_changes)
        _print_summary(results)
    elif mode == 'incomplete':
        _print_result(results, results.names.all_incomplete)
    elif mode == 'fixes':
        _print_result(results, results.names.all_fixes)
    elif mode == 'problems':
        _print_result(results, results.names.all_problems)
    elif mode == 'regressions':
        _print_result(results, results.names.all_regressions)
    elif mode == 'summary':
        _print_summary(results)



# coding=utf-8
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" Module provides a base class for Tests """

import abc
import copy
import errno
import itertools
import os
import signal
import subprocess
import sys
import time
import traceback
import warnings

from framework import exceptions
from framework import status
from framework.options import OPTIONS
from framework.results import TestResult

_EXTRA_POPEN_ARGS = {}

if os.name == 'posix':
    # This should work for all *nix systems, Linux, the BSDs, and OSX.
    # This speicifically creates a session group for each test, so that
    # it's children can be killed if it times out.
    _EXTRA_POPEN_ARGS = {'start_new_session': True}

# pylint: enable=wrong-import-position,wrong-import-order


__all__ = [
    'Test',
    'TestIsSkip',
    'TestRunError',
    'ValgrindMixin',
    'WindowResizeMixin',
    'is_crash_returncode',
]

# Allows timeouts to be suppressed by setting the environment variable
# PIGLIT_NO_TIMEOUT to anything that bool() will resolve as True
_SUPPRESS_TIMEOUT = bool(os.environ.get('PIGLIT_NO_TIMEOUT', False))


class TestIsSkip(exceptions.PiglitException):
    """Exception raised in is_skip() if the test is a skip."""
    def __init__(self, reason):
        super(TestIsSkip, self).__init__()
        self.reason = reason


class TestRunError(exceptions.PiglitException):
    """Exception raised if the test fails to run."""
    def __init__(self, message, status):
        super(TestRunError, self).__init__(message)
        self.status = status


def is_crash_returncode(returncode):
    """Determine whether the given process return code correspond to a
    crash.
    """
    # In Python 3 isn't allowed to compare NoneType and other types.
    if returncode is None:
        return False

    if sys.platform == 'win32':
        # On Windows:
        # - For uncaught exceptions the process terminates with the exception
        # code, which is usually negative
        # - MSVCRT's abort() terminates process with exit code 3
        return returncode < 0 or returncode == 3
    else:
        return returncode < 0


class Test(metaclass=abc.ABCMeta):
    """ Abstract base class for Test classes

    This class provides the framework for running tests, with several methods
    and properties that can be overwritten to produce a specialized class for
    running test suites other than piglit.

    It provides two methods for running tests, execute and run.
    execute() provides lots of features, and is invoked when running piglit
    from the command line, run() is a more basic method for running the test,
    and is called internally by execute(), but is can be useful outside of it.

    Arguments:
    command -- a value to be passed to subprocess.Popen

    Keyword Arguments:
    run_concurrent -- If True the test is thread safe. Default: False

    """
    __slots__ = ['run_concurrent', 'env', 'result', 'cwd', '_command']
    timeout = None

    def __init__(self, command, run_concurrent=False, env=None, cwd=None):
        assert isinstance(command, list), command

        self.run_concurrent = run_concurrent
        self._command = copy.copy(command)
        self.env = env or {}
        self.result = TestResult()
        self.cwd = cwd

    def execute(self, path, log, options):
        """ Run a test

        Run a test, but with features. This times the test, uses dmesg checking
        (if requested), and runs the logger.

        Arguments:
        path    -- the name of the test
        log     -- a log.Log instance
        options -- a dictionary containing dmesg and monitoring objects
        """
        log.start(path)
        # Run the test
        if OPTIONS.execute:
            try:
                self.result.time.start = time.time()
                options['dmesg'].update_dmesg()
                options['monitor'].update_monitoring()
                self.run()
                self.result.time.end = time.time()
                self.result = options['dmesg'].update_result(self.result)
                options['monitor'].check_monitoring()
            # This is a rare case where a bare exception is okay, since we're
            # using it to log exceptions
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exc(file=sys.stderr)
                self.result.result = 'fail'
                self.result.exception = "{}{}".format(exc_type, exc_value)
                self.result.traceback = "".join(
                    traceback.format_tb(exc_traceback))

            log.log(self.result.result)
        else:
            log.log('dry-run')

    @property
    def command(self):
        assert self._command
        return self._command

    @command.setter
    def command(self, new):
        assert isinstance(new, list), 'Test.command must be a list'
        self._command = new

    @abc.abstractmethod
    def interpret_result(self):
        """Convert the raw output of the test into a form piglit understands.
        """
        if is_crash_returncode(self.result.returncode):
            self.result.result = status.CRASH
            if self.result.subtests:
                # We know because subtests are ordered that the first test with
                # a status of NOTRUN is the subtest that crashed, mark that
                # test and move on.
                for k, v in self.result.subtests.items():
                    if v == status.NOTRUN:
                        self.result.subtests[k] = status.CRASH
                        break
        elif self.result.returncode != 0:
            if self.result.result == status.PASS:
                self.result.result = status.WARN
            else:
                self.result.result = status.FAIL

    def run(self):
        """
        Run a test.  The return value will be a dictionary with keys
        including 'result', 'info', 'returncode' and 'command'.
        * For 'result', the value may be one of 'pass', 'fail', 'skip',
          'crash', or 'warn'.
        * For 'info', the value will include stderr/out text.
        * For 'returncode', the value will be the numeric exit code/value.
        * For 'command', the value will be command line program and arguments.
        """
        self.result.command = ' '.join(map(str, self.command))
        self.result.environment = " ".join(
            '{0}="{1}"'.format(k, v) for k, v in itertools.chain(
                OPTIONS.env.items(), self.env.items()))

        try:
            self.is_skip()
        except TestIsSkip as e:
            self.result.result = status.SKIP
            for each in self.result.subtests.keys():
                self.result.subtests[each] = status.SKIP
            self.result.out = e.reason
            self.result.returncode = None
            return

        try:
            self._run_command()
        except TestRunError as e:
            self.result.result = str(e.status)
            for each in self.result.subtests.keys():
                self.result.subtests[each] = str(e.status)
            self.result.out = str(e)
            self.result.returncode = None
            return

        self.interpret_result()

    def is_skip(self):
        """ Application specific check for skip

        If this function returns a truthy value then the current test will be
        skipped. The base version will always return False

        """
        pass

    def _run_command(self, **kwargs):
        """ Run the test command and get the result

        This method sets environment options, then runs the executable. If the
        executable isn't found it sets the result to skip.

        """
        # This allows the ReducedProcessMixin to work without having to whack
        # self.command (which should be treated as immutable), but is
        # considered private.
        command = kwargs.pop('_command', self.command)

        # Setup the environment for the test. Environment variables are taken
        # from the following sources, listed in order of increasing precedence:
        #
        #   1. This process's current environment.
        #   2. Global test options. (Some of these are command line options to
        #      Piglit's runner script).
        #   3. Per-test environment variables set in all.py.
        #
        # Piglit chooses this order because Unix tradition dictates that
        # command line options (2) override environment variables (1); and
        # Piglit considers environment variables set in all.py (3) to be test
        # requirements.
        _base = itertools.chain(os.environ.items(),
                                OPTIONS.env.items(),
                                self.env.items())
        fullenv = {str(k): str(v) for k, v in _base}

        try:
            proc = subprocess.Popen(map(str, command),
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    cwd=self.cwd,
                                    env=fullenv,
                                    universal_newlines=True,
                                    **_EXTRA_POPEN_ARGS)

            self.result.pid.append(proc.pid)
            if not _SUPPRESS_TIMEOUT:
                out, err = proc.communicate(timeout=self.timeout)
            else:
                out, err = proc.communicate()
            returncode = proc.returncode
        except OSError as e:
            # Different sets of tests get built under different build
            # configurations.  If a developer chooses to not build a test,
            # Piglit should not report that test as having failed.
            if e.errno == errno.ENOENT:
                raise TestRunError("Test executable not found.\n", 'skip')
            else:
                raise e
        except subprocess.TimeoutExpired:
            # This can only be reached if subprocess32 is present or on python
            # 3.x, since # TimeoutExpired is never raised by the python 2.7
            # fallback code.

            proc.terminate()

            if sys.platform != 'win32':
                if proc.poll() is None:
                    time.sleep(3)
                    os.killpg(os.getpgid(proc.pid), signal.SIGKILL)

            # Since the process isn't running it's safe to get any remaining
            # stdout/stderr values out and store them.
            self.result.out, self.result.err = proc.communicate()

            raise TestRunError(
                'Test run time exceeded timeout value ({} seconds)\n'.format(
                    self.timeout),
                'timeout')
        # LLVM prints colored text into stdout/stderr on error, which raises:
        except UnicodeDecodeError as e:
            raise TestRunError("UnicodeDecodeError.\n", 'crash')

        # The setter handles the bytes/unicode conversion
        self.result.out = out
        self.result.err = err
        self.result.returncode = returncode

    def __eq__(self, other):
        return self.command == other.command

    def __ne__(self, other):
        return not self == other


class DummyTest(Test):
    def __init__(self, name, result):
        super(DummyTest, self).__init__([name])
        self.result.result = result

    def execute(self, path, log, options):
        pass

    def interpret_result(self):
        pass


class WindowResizeMixin(object):
    """ Mixin class that deals with spurious window resizes

    On gnome (and possible other DE's) the window manager may decide to resize
    a window. This causes the test to fail even though otherwise would not.
    This Mixin overrides the _run_command method to run the test 5 times, each
    time searching for the string 'Got spurious window resize' in the output,
    if it fails to find it it will break the loop and continue.

    see: https://bugzilla.gnome.org/show_bug.cgi?id=680214

    """
    def _run_command(self, *args, **kwargs):
        """Run a test up 5 times when window resize is detected.

        Rerun the command up to 5 times if the window size changes, if it
        changes 6 times mark the test as fail and return True, which will cause
        Test.run() to return early.

        """
        for _ in range(5):
            super(WindowResizeMixin, self)._run_command(*args, **kwargs)
            if "Got spurious window resize" not in self.result.out:
                return

        # If we reach this point then there has been no error, but spurious
        # resize was detected more than 5 times. Set the result to fail
        raise TestRunError('Got spurious resize more than 5 times', 'fail')


class ValgrindMixin(object):
    """Mixin class that adds support for running tests through valgrind.

    This mixin allows a class to run with the --valgrind option.

    """
    @Test.command.getter
    def command(self):
        return super(ValgrindMixin, self).command

    def keys(self):
        if OPTIONS.valgrind:
            return ['valgrind', '--quiet', '--error-exitcode=1',
                    '--tool=memcheck']
        else:
            return []

    def interpret_result(self):
        """Set the status to the valgrind status.

        It is important that the valgrind interpret_results code is run last,
        since it depends on the statuses already set and passed to it,
        including the Test.interpret_result() method. To this end it executes
        super().interpret_result(), then calls it's own result.

        """
        super(ValgrindMixin, self).interpret_result()

        if OPTIONS.valgrind:
            # If the underlying test failed, simply report
            # 'skip' for this valgrind test.
            if self.result.result != 'pass' and (
                    self.result.result != 'warn' and
                    self.result.returncode != 0):
                self.result.result = 'skip'
            elif self.result.returncode == 0:
                # Test passes and is valgrind clean.
                self.result.result = 'pass'
            else:
                # Test passed but has valgrind errors.
                self.result.result = 'fail'


class ReducedProcessMixin(metaclass=abc.ABCMeta):
    """This Mixin simplifies writing Test classes that run more than one test
    in a single process.

    Although one of the benefits of piglit is it's process isolation, there are
    times that process isolation is too expensive for day to day runs, and
    running more than one test in a single process is a valid trade-off for
    decreased run times. This class helps to ease writing a Test class for such
    a purpose, while not suffering all of the drawback of the approach.

    The first way that this helps is that it provides crash detection and
    recovery, allowing a single subtest to crash
    """

    def __init__(self, command, subtests=None, **kwargs):
        assert subtests is not None
        super(ReducedProcessMixin, self).__init__(command, **kwargs)
        self._expected = subtests
        self._populate_subtests()

    def is_skip(self):
        """Skip if the length of expected is 0."""
        if not self._expected:
            raise TestIsSkip('All subtests skipped')
        super(ReducedProcessMixin, self).is_skip()

    def __find_sub(self):
        """Helper for getting the next index."""
        return len([l for l in self.result.out.split('\n')
                    if self._is_subtest(l)])

    @staticmethod
    def _subtest_name(test):
        """If the name provided isn't the subtest name, this method does."""
        return test

    def _stop_status(self):
        """This method returns the status of the test that stopped the run.

        By default this will return status.CRASH, but this may not be suitable
        for some suites, which may require special considerations and need to
        require a different status in some cases, like SKIP.
        """
        return status.CRASH

    def _run_command(self, *args, **kwargs):
        """Run the command until all of the subtests have completed or crashed.

        This method will try to run all of the subtests, resuming the run if
        it's interrupted, and combining the stdout and stderr attributes
        together for parsing later. I will separate those values with
        "\n\n====RESUME====\n\n".
        """
        super(ReducedProcessMixin, self)._run_command(*args, **kwargs)

        if not self._is_cherry():
            returncode = self.result.returncode
            out = [self.result.out]
            err = [self.result.err]
            cur_sub = self.__find_sub() or 1
            last = len(self._expected)

            while cur_sub < last:
                self.result.subtests[
                    self._subtest_name(self._expected[cur_sub - 1])] = \
                        self._stop_status()

                super(ReducedProcessMixin, self)._run_command(
                    _command=self._resume(cur_sub) + list(args), **kwargs)

                out.append(self.result.out)
                err.append(self.result.err)

                # If the index is 0 the next test failed without printing a
                # name, increase by 1 so that test will be marked crash and we
                # don't get stuck in an infinite loop, otherwise return the
                # number of tests that did complete.
                cur_sub += self.__find_sub() or 1

            if not self._is_cherry():
                self.result.subtests[
                    self._subtest_name(self._expected[cur_sub - 1])] = \
                        self._stop_status()

            # Restore and keep the original returncode (so that it remains a
            # non-pass, since only one test might fail and the resumed part
            # might return 0)
            self.result.returncode = returncode
            self.result.out = '\n\n====RESUME====\n\n'.join(out)
            self.result.err = '\n\n====RESUME====\n\n'.join(err)

    def _is_cherry(self):
        """Method used to determine if rerunning is required.

        If this returns False then the rerun path will be entered, otherwise
        _run_command is effectively a bare call to super().

        Classes using this mixin may need to overwrite this if the binary
        they're calling can stop prematurely but return 0.
        """
        return self.result.returncode == 0

    def _populate_subtests(self):
        """Default implementation of subtest prepopulation.

        It may be necissary to override this depending on the subtest format.
        """
        self.result.subtests.update({x: status.NOTRUN for x in self._expected})

    @abc.abstractmethod
    def _resume(self, current):
        """Method that defines how to resume the case if it crashes.

        This method will be provided with a completed count, which is the index
        into self._expected of the first subtest that hasn't been run. This
        method should return the command to restart, and the ReduceProcessMixin
        will handle actually restarting the the process with the new command.
        """

    @abc.abstractmethod
    def _is_subtest(self, line):
        """Determines if a line in stdout contains a subtest name.

        This method is used during the resume detection phase of the
        _run_command method to determine how many subtests have successfully
        been run.

        Should simply return True if the line represents a test starting, or
        False if it does not.
        """



# coding=utf-8
# Copyright (c) 2013-2016, 2019 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" Module for terminal logging

This module provides a class LogManager, which acts as a state manager
returning BaseLog derived instances to individual tests.

"""

import sys
import abc
import itertools
import threading
import collections
from http.server import HTTPServer, BaseHTTPRequestHandler

import json

from framework.core import PIGLIT_CONFIG
from framework import grouptools

__all__ = ['LogManager']


class BaseLog(metaclass=abc.ABCMeta):
    """ Abstract base class for Log objects

    It provides a lock, which should be used to lock whenever the shared state is
    modified, or when printing to the screen (or both).

    Arguments:
    state -- the state dict from LogManager

    """
    SUMMARY_KEYS = set([
        'pass', 'fail', 'warn', 'crash', 'skip', 'dmesg-warn', 'dmesg-fail',
        'dry-run', 'timeout'])
    _LOCK = None

    def __init__(self, state, state_lock):
        self._LOCK = state_lock
        self._state = state
        self._pad = len(str(state['total']))

    @abc.abstractmethod
    def start(self, name):
        """ Called before test run starts

        This method is used to print things before the test starts

        """

    @abc.abstractmethod
    def log(self, status):
        """ Print the result of the test

        This method is run after the test completes, and is used to log the
        actual result of the test

        """

    @abc.abstractmethod
    def summary(self):
        """ Print a final summary

        This method is run at the end of the test run, and is used to print a
        final summary of the run

        """


class QuietLog(BaseLog):
    """ A logger providing minimal status output

    This logger provides only a ninja like [x/y] pass: z, fail: w ... output.

    It uses \r to print over the same line when printing to a tty. If
    sys.stdout is not a tty then it prints \n at the end of the line instead

    Arguments:
    status -- the status to print

    """
    _test_counter = itertools.count()

    def __init__(self, *args, **kwargs):
        super(QuietLog, self).__init__(*args, **kwargs)

        # If the output is a tty we can use '\r' (return, no linebreak) to
        # print over the existing line, if stdout isn't a tty that will not
        # work (and can break systems like jenkins), so we print a \n instead
        if sys.stdout.isatty():
            self._endcode = '\r'
        else:
            self._endcode = '\n'

        self.__counter = next(self._test_counter)

    def start(self, name):
        # This cannot be done in the constructor, since the constructor gets
        # called for the final summary too.
        with self._LOCK:
            self._state['running'].append(self.__counter)

    def _log(self, status):
        """ non-locked helper for logging

        To allow for code sharing with a subclass using a non-recursive lock we
        need to share this code in a an unlocked form.

        """
        # increment complete
        self._state['complete'] += 1

        # Add to the summary dict
        assert status in self.SUMMARY_KEYS, \
            'Invalid status for logger: {}'.format(status)
        self._state['summary'][status] += 1

        self._print_summary()
        self._state['running'].remove(self.__counter)

    def log(self, status):
        with self._LOCK:
            self._log(status)

    def summary(self):
        with self._LOCK:
            self._print_summary()
            self._print('\n')

    def _print_summary(self):
        """ Print the summary result

        this prints '[done/total] {status}', it is a private method hidden from
        the children of this class (VerboseLog)

        """
        assert self._LOCK.locked()

        out = '[{done}/{total}] {status} {running}'.format(
            done=str(self._state['complete']).zfill(self._pad),
            total=str(self._state['total']).zfill(self._pad),
            status=', '.join('{0}: {1}'.format(k, v) for k, v in
                             sorted(self._state['summary'].items())),
            running=''.join('|/-\\'[x % 4] for x in self._state['running'])
        )

        self._print(out)

    def _print(self, out):
        """ Shared print method that ensures any bad lines are overwritten """
        assert self._LOCK.locked()

        # If the new line is shorter than the old one add trailing
        # whitespace
        pad = self._state['lastlength'] - len(out)

        sys.stdout.write(out)
        if pad > 0:
            sys.stdout.write(' ' * pad)
        sys.stdout.write(self._endcode)
        sys.stdout.flush()

        # Set the length of the line just printed (minus the spaces since
        # we don't care if we leave whitespace) so that the next line will
        # print extra whitespace if necissary
        self._state['lastlength'] = len(out)


class VerboseLog(QuietLog):
    """ A more verbose version of the QuietLog

    This logger works like the quiet logger, except it doesn't overwrite the
    previous line, ever. It also prints an additional line at the start of each
    test, which the quite logger doesn't do.

    """
    def __init__(self, *args, **kwargs):
        super(VerboseLog, self).__init__(*args, **kwargs)
        self.__name = None    # the name needs to be printed by start and log

    def _print(self, out, newline=False):
        """ Print to the console, add a newline if necissary

        For the verbose logger there are times that one wants both an
        overwritten line, and a line that is static. This method adds the
        ability to print a newline character at the end of the line.

        This is useful for the non-status lines (running: <name>, and <status>:
        <name>), since these lines should be be overwritten, but the running
        status line should.

        """
        super(VerboseLog, self)._print(grouptools.format(out))
        if newline:
            sys.stdout.write('\n')
            sys.stdout.flush()

    def start(self, name):
        """ Print the test that is being run next

        This printings a running: <testname> message before the test run
        starts.

        """
        super(VerboseLog, self).start(name)
        with self._LOCK:
            self._print('running: {}'.format(name), newline=True)
            self.__name = name
            self._print_summary()

    def _log(self, value):
        """ Print a message after the test finishes

        This method prints <status>: <name>. It also does a little bit of magic
        before calling super() to print the status line.

        """
        self._print('{0}: {1}'.format(value, self.__name), newline=True)

        # Set lastlength to 0, this prevents printing needless padding in
        # super()
        self._state['lastlength'] = 0
        super(VerboseLog, self)._log(value)


class DummyLog(BaseLog):
    """ A Logger that does nothing """
    def __init__(self, state, state_lock):
        pass

    def start(self, name):
        pass

    def log(self, status):
        pass

    def summary(self):
        pass


class HTTPLogServer(threading.Thread):
    class RequestHandler(BaseHTTPRequestHandler):
        INDENT = 4

        def do_GET(self):
            if self.path == "/summary":
                self.send_response(200)
                self.end_headers()
                with self.server.state_lock:
                    status = {
                        "complete": self.server.state["complete"],
                        "running" : self.server.state["running"],
                        "total"   : self.server.state["total"],
                        "results" : self.server.state["summary"],
                    }
                self.wfile.write(json.dumps(status, indent=self.INDENT))
            else:
                self.send_response(404)
                self.end_headers()

    def __init__(self, state, state_lock):
        super(HTTPLogServer, self).__init__()
        port = int(PIGLIT_CONFIG.safe_get("http", "port", fallback=8080))
        self._httpd = HTTPServer(("", port), HTTPLogServer.RequestHandler)
        self._httpd.state = state
        self._httpd.state_lock = state_lock

    def run(self):
        while True:
            with self._httpd.state_lock:
                # stop handling requests after the request for the final results
                if self._httpd.state["complete"] == self._httpd.state["total"]:
                    break;
            self._httpd.handle_request()


class HTTPLog(BaseLog):
    """ A Logger that serves status information over http """

    def __init__(self, state, state_lock):
        super(HTTPLog, self).__init__(state, state_lock)
        self._name = None

    def start(self, name):
        with self._LOCK:
            self._name = name
            self._state['running'].append(self._name)

    def log(self, status):
        with self._LOCK:
            self._state['running'].remove(self._name)
            self._state['complete'] += 1
            assert status in self.SUMMARY_KEYS
            self._state['summary'][str(status)] += 1

    def summary(self):
        pass


class LogManager(object):
    """ Creates new log objects

    Each of the log objects it creates have two accessible methods: start() and
    log(); neither method should return anything, and they must be thread safe.
    log() should accept a single argument, which is the status the test
    returned. start() should also take a single argument, the name of the test

    When the LogManager is initialized it is initialized with an argument which
    determines which Log class it will return (they are set via the LOG_MAP
    dictionary, which maps string names to class callables), it will return
    these as long as it exists.

    Arguments:
    logger -- a string name of a logger to use
    total -- the total number of test to run

    """
    LOG_MAP = {
        'quiet': QuietLog,
        'verbose': VerboseLog,
        'dummy': DummyLog,
        'http': HTTPLog,
    }

    def __init__(self, logger, total):
        assert logger in self.LOG_MAP
        self._log = self.LOG_MAP[logger]
        self._state = {
            'total': total,
            'summary': collections.defaultdict(lambda: 0),
            'lastlength': 0,
            'complete': 0,
            'running': [],
        }
        self._state_lock = threading.Lock()

        # start the http server for http logger
        if logger == 'http':
            self.log_server = HTTPLogServer(self._state, self._state_lock)
            self.log_server.start()

    def get(self):
        """ Return a new log instance """
        return self._log(self._state, self._state_lock)



# coding=utf-8
# Copyright (c) 2016, 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import json

from framework import profile, status


class FeatResults(object):  # pylint: disable=too-few-public-methods
    """Container object for results.

    Has the results, feature profiles and feature computed results.

    """
    def __init__(self, results, json_file):

        with open(json_file) as data:
            feature_data = json.load(data)

        self.feat_fractions = {}
        self.feat_status = {}
        self.features = set()
        self.results = results

        profiles = {}

        # we expect all the result sets to be for the same profile
        profile_orig = profile.load_test_profile(results[0].options['profile'][0])

        for feature in feature_data:
            self.features.add(feature)

            profiles[feature] = profile_orig.copy()

            incl_str = feature_data[feature]["include_tests"]
            excl_str = feature_data[feature]["exclude_tests"]

            profiles[feature].filters.append(
                profile.RegexFilter(
                    [incl_str] if incl_str and not incl_str.isspace() else []))
            profiles[feature].filters.append(
                profile.RegexFilter(
                    [excl_str] if excl_str and not excl_str.isspace() else [],
                    inverse=True))

        for results in self.results:
            self.feat_fractions[results.name] = {}
            self.feat_status[results.name] = {}

            for feature in feature_data:
                result_set = set(results.tests)
                profile_set = set(a for a, _ in profiles[feature].itertests())

                common_set = profile_set & result_set
                passed_list = [x for x in common_set if results.tests[x].result == status.PASS]

                total = len(common_set)
                passed = len(passed_list)

                self.feat_fractions[results.name][feature] = (passed, total)
                if total == 0:
                    self.feat_status[results.name][feature] = status.NOTRUN
                else:
                    if 100 * passed // total >= feature_data[feature]["target_rate"]:
                        self.feat_status[results.name][feature] = status.PASS
                    else:
                        self.feat_status[results.name][feature] = status.FAIL



# coding=utf-8
#
# Copyright © 2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

import argparse

from framework import exceptions
from framework.replay import download_utils, options

from . import parsers

__all__ = ['download']


def _ensure_file(args):
    options.OPTIONS.set_download_url(args.download_url)
    options.OPTIONS.set_download_caching_proxy_url(args.download_caching_proxy_url)
    options.OPTIONS.download['force'] = args.force_download
    options.OPTIONS.download['minio_host'] = args.download_minio_host
    options.OPTIONS.download['minio_bucket'] = args.download_minio_bucket
    options.OPTIONS.download['role_session_name'] = args.download_role_session_name
    options.OPTIONS.download['jwt'] = args.download_jwt
    options.OPTIONS.db_path = args.db_path

    return download_utils.ensure_file(args.file_path)


@exceptions.handler
def download(input_):
    """ Parser for replayer download command """
    parser = argparse.ArgumentParser(parents=[parsers.DOWNLOAD_URL,
                                              parsers.DOWNLOAD_CACHING_PROXY_URL,
                                              parsers.DOWNLOAD_FORCE,
                                              parsers.DOWNLOAD_MINIO_HOST,
                                              parsers.DOWNLOAD_MINIO_BUCKET,
                                              parsers.DOWNLOAD_ROLE_SESSION_NAME,
                                              parsers.DOWNLOAD_JWT,
                                              parsers.DB_PATH])
    parser.add_argument(
        'file_path',
        help=('the path to the file '
              'at the provided URL from which to download'))
    parser.set_defaults(func=_ensure_file)

    args = parser.parse_args(input_)

    args.func(args)



# coding=utf-8
#
# Copyright (c) 2014, 2016, 2019 Intel Corporation
# Copyright © 2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT


""" Base class for dump backends

This module provides a base class for replayer dump backend modules.

"""

import abc
import functools
import subprocess
import sys
from os import path

from framework import core
from framework.replay.options import OPTIONS


class DumpBackend(metaclass=abc.ABCMeta):
    """ Base class for dump backends

    This class provides an basic ancestor for classes implementing dump
    backends, providing a light public API. The goal of this API is to be "just
    enough", not a generic writing solution. To that end it provides the
    method, 'dump'. This method is designed to be just enough to write a
    backend without needing format specific options.

    """
    def __init__(self, trace_path, output_dir=None, calls=None, **kwargs):
        """ Generic constructor

        This method takes keyword arguments that define options for the
        backends. Options should be prefixed to identify which backends they
        apply to. For example, an apitrace specific value should be passed as
        apitrace_*, while a file gfxrecon value should be passed as gfxrecon_*)

        Arguments:

        trace_path -- the path to the trace from which we want to dump calls as
                      images.
        output_dir -- the place to write the images to.
        calls      -- an array of the calls in the trace for which we want to
                      dump images.

        """
        self._trace_path = trace_path
        self._output_dir = output_dir
        self._calls = calls or []

        if self._output_dir is None:
            self._output_dir = path.join('trace', OPTIONS.device_name,
                                         path.dirname(self._trace_path))

    @staticmethod
    def log(severity, msg, end='\n'):
        print('[dump_trace_images] {}: {}'.format(severity, msg), flush=True,
              end=end)

    @staticmethod
    def log_result(msg):
        print(msg, flush=True)

    @staticmethod
    def _run_logged_command(cmd, env):
        # Explicitly send the stderr to the fd at sys.stderr in case it was
        # redirected for the parent process.
        # See:
        # https://bugs.python.org/issue44158
        ret = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=sys.stderr,
                             env=env)
        logoutput = '[dump_trace_images] Running: {}\n'.format(
            ' '.join(cmd)).encode() + ret.stdout
        print(logoutput.decode(errors='replace'))
        if ret.returncode:
            raise RuntimeError(
                '[dump_trace_images] Process failed with error code: {}'.format(
                    ret.returncode))

    @abc.abstractmethod
    def _get_last_frame_call(self):
        """Get the number of the last frame call from the trace"""

    @abc.abstractmethod
    def dump(self):
        """ Dump the calls to images from the trace

        This method actually dumps the calls from a trace.

        """


def dump_handler(func):
    """ Decorator function for handling trace dumps.

    This will handle exceptions and log the result.

    """

    @functools.wraps(func)
    def _inner(self, *args, **kwargs):
        try:
            DumpBackend.log('Info',
                            'Dumping trace {}'.format(self._trace_path),
                            end='...\n')
            core.check_dir(self._output_dir)
            func(self, *args, **kwargs)
            DumpBackend.log_result('OK')
            return True
        except Exception as e:
            DumpBackend.log_result('ERROR')
            DumpBackend.log('Debug', '=== Failure log start ===')
            print(e)
            DumpBackend.log('Debug', '=== Failure log end ===')
            return False

    return _inner



# coding=utf-8
# Copyright (c) 2014, 2016, 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


""" Base classes for backends

This module provides mixins and base classes for backend modules.

"""

import abc
import contextlib
import itertools
import os
import shutil

from framework import options
from . import compression
from framework.results import TestResult
from framework.status import INCOMPLETE


@contextlib.contextmanager
def write_compressed(filename):
    """Write a the final result using desired compression.

    This helper function reads the piglit.conf to decide whether to use
    compression, and what type of compression to use.

    Currently it implements no compression

    """
    mode = compression.get_mode()
    if mode != 'none':
        # if the suffix (final .xxx) is a known compression suffix
        suffix = os.path.splitext(filename)[1]
        if suffix in compression.COMPRESSION_SUFFIXES:
            filename = '{}.{}'.format(os.path.splitext(filename)[0], mode)
        else:
            filename = '{}.{}'.format(filename, mode)

    with compression.COMPRESSORS[mode](filename) as f:
        yield f


class Backend(metaclass=abc.ABCMeta):
    """ Abstract base class for summary backends

    This class provides an abstract ancestor for classes implementing backends,
    providing a light public API. The goal of this API is to be "just enough",
    not a generic writing solution. To that end it provides two public methods,
    'finalize', and 'write_test'. These two methods are designed to be just
    enough to write a backend without needing format specific options.

    Any locking that is necessary should be done in the child classes, as not
    all potential backends need locking (for example, a SQL based backend might
    be thread safe and not need to be locked during write)

    """
    @abc.abstractmethod
    def __init__(self, dest, metadata, **kwargs):
        """ Generic constructor

        This method should setup the container and open any files or connections
        as necessary. It should not however, write anything into the backend
        store, that job is for the initialize method.

        In addition it takes keyword arguments that define options for the
        backends. Options should be prefixed to identify which backends they
        apply to. For example, a json specific value should be passed as
        json_*, while a file specific value should be passed as file_*)

        Arguments:
        dest -- the place to write the results to. This should be correctly
                handled based on the backend, the example is calls open() on a
                file, but other backends might want different options

        """

    @abc.abstractmethod
    def initialize(self, metadata):
        """ Write initial metadata and setup

        This method is used to write metadata into the backend store and do any
        other initial backend writing that is required. This method and the
        finalize() method are bookends, one starts, the other finishes.

        Arguments:
        metadata -- a dict or dict-like object that contains metadata to be
                    written into the backend

        """

    @abc.abstractmethod
    def finalize(self, metadata=None):
        """ Write final metadata into to the store and close it

        This method writes any final metadata into the store, what can be
        written is implementation specific, backends are free to ignore any
        data that is not applicable.

        metadata is not required, and Backend derived classes need to handle
        being passed None correctly.

        Keyword Arguments:
        metadata -- Any metadata to be written in after the tests, should be a
                    dict or dict-like object


        """

    @abc.abstractmethod
    def write_test(self, name):
        """ Write a test into the backend store

        This method writes an actual test into the backend store.

        Should be a context manager, used with the with statement. It should
        first write an incomplete status value, then yield and object that will
        overwrite that value with the final value. That object needs to take a
        'data' parameter which is a result.TestResult object.

        Arguments:
        name -- the name of the test to be written
        data -- a TestResult object representing the test data

        """


class FileBackend(Backend):
    """ A baseclass for file based backends

    This class provides a few methods and setup required for a file based
    backend.

    Arguments:
    dest -- a folder to store files in

    Keyword Arguments:
    file_start_count -- controls the counter for the test result files.
                        Whatever this is set to will be the first name of the
                        tests. It is important for resumes that this is not
                        overlapping as the Inheriting classes assume they are
                        not. Default: 0

    """
    def __init__(self, dest, file_start_count=0, **kwargs):
        self._dest = dest
        self._counter = itertools.count(file_start_count)
        self._write_final = write_compressed

    __INCOMPLETE = TestResult(result=INCOMPLETE)

    def __fsync(self, file_):
        """ Sync the file to disk

        If options.OPTIONS.sync is truthy this will sync self._file to disk

        """
        file_.flush()
        if options.OPTIONS.sync:
            os.fsync(file_.fileno())

    @abc.abstractmethod
    def _write(self, f, name, data):
        """Method that writes a TestResult into a result file."""

    @abc.abstractproperty
    def _file_extension(self):
        """The file extension of the backend."""

    @contextlib.contextmanager
    def write_test(self, name):
        """Write a test.

        When this context manager is opened it will first write a placeholder
        file with the status incomplete.

        When it is called to write the final result it will create a temporary
        file, write to that file, then move that file over the original,
        incomplete status file. This helps to make the operation atomic, as
        long as the filesystem continues running and the result was valid in
        the original file it will be valid at the end

        """
        def finish(val):
            tfile = file_ + '.tmp'
            with open(tfile, 'w') as f:
                self._write(f, name, val)
                self.__fsync(f)
            shutil.move(tfile, file_)

        file_ = os.path.join(self._dest, 'tests', '{}.{}'.format(
            next(self._counter), self._file_extension))

        with open(file_, 'w') as f:
            self._write(f, name, self.__INCOMPLETE)
            self.__fsync(f)

        yield finish



# coding=utf-8
#
# Copyright (c) 2020 Collabora Ltd
# Copyright © 2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

import json  # type: ignore
import os
import shutil
from glob import glob
from os import path

from framework import core, status
from framework.replay import backends
from framework.replay import query_traces_yaml as qty
from framework.replay.download_utils import ensure_file
from framework.replay.image_checksum import hexdigest_from_image
from framework.replay.options import OPTIONS

__all__ = ['from_yaml',
           'trace']


def _replay(trace_path, results_path):
    try:
        success = backends.dump(trace_path, results_path, [])
    except (backends.DumpBackendNotImplementedError,
            backends.DumpBackendError) as e:
        print(e)
        success = False

    if not success:
        print("[check_image] Trace {} couldn't be replayed. "
              "See above logs for more information.".format(trace_path))
        return None, None
    else:
        file_name = path.basename(trace_path)
        files = glob(path.join(results_path, file_name + '-*' + '.png'))
        if not files:
            print('[check_image] No dumped files found '
                  'in the results path "{}". '
                  'See above logs for more information.'.format(results_path))
            return None, None
        image_file = files[0]
        return hexdigest_from_image(image_file), image_file


def _check_trace(trace_path, expected_checksum):
    ensure_file(trace_path)

    json_result = {}

    trace_dir = path.dirname(trace_path)
    dir_in_results = path.join('trace', OPTIONS.device_name or '', trace_dir)
    results_path = path.join(OPTIONS.results_path, dir_in_results)
    core.check_dir(results_path)

    checksum, image_file = _replay(path.join(OPTIONS.db_path, trace_path),
                                   results_path)

    print('[check_image]\n'
          '    actual: {}\n'
          '  expected: {}'.format(checksum or 'error', expected_checksum))

    json_result['images'] = [
        {'image_desc': trace_path,
         'checksum_ref': expected_checksum,
         'checksum_render': None,
         'image_ref': expected_checksum + '.png',
         'image_render': None}]

    if checksum is None:
        return status.CRASH, json_result

    json_result['images'][0]['checksum_render'] = checksum
    json_result['images'][0]['image_render'] = checksum + '.png'

    if checksum == expected_checksum:
        if not OPTIONS.keep_image:
            os.remove(image_file)
        print('[check_image] Images match for:\n  {}\n'.format(trace_path))
        result = status.PASS
    else:
        print('[check_image] Images differ for:\n  {}'.format(trace_path))
        print('[check_image] For more information see '
              'https://gitlab.freedesktop.org/'
              'mesa/piglit/blob/master/replayer/README.md\n')
        result = status.FAIL

    if result is not status.PASS or OPTIONS.keep_image:
        root, ext = path.splitext(image_file)
        image_file_dest = '{}-{}{}'.format(root, checksum, ext)
        shutil.move(image_file, image_file_dest)
        json_result['images'][0]['image_render'] = image_file_dest

    return result, json_result


def _print_result(result, trace_path, json_result):
    output = 'PIGLIT: '
    json_result['result'] = str(result)

    output += json.dumps(json_result)
    print(output)


def from_yaml(yaml_file):
    y = qty.load_yaml(yaml_file)

    OPTIONS.set_download_url(qty.download_url(y))

    global_result = status.PASS
    # TODO: print in subtest format
    # json_results = {}
    t_list = qty.traces(y, device_name=OPTIONS.device_name, checksum=True)
    for t in t_list:
        result, json_result = _check_trace(t['path'], t['checksum'])
        if result is not status.PASS and global_result is not status.CRASH:
            global_result = result
        # json_results.update(json_result)
        # _print_result(result, t['path'], json_result)

    return global_result


def trace(trace_path, expected_checksum):
    result, json_result = _check_trace(trace_path, expected_checksum)
    _print_result(result, trace_path, json_result)

    return result



#!/usr/local/bin/python3.13
# coding=utf-8
#
# Copyright (c) 2014, 2019 Intel Corporation
# Copyright © 2020 Valve Corporation.
# Copyright © 2021 Collabora Ltd.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT


""" Wrapper for replayer executable

This imports functions from the replay framework and calls them with the
argument parts that the parser defined here doesn't know how to parse.

It is very important that the final parser not generate a help message
(add_help=False in the constructor arguments), otherwise this parser will
capture -h/--help and the results will not be useful.

"""

import argparse
import os
import os.path as path
import sys


def setup_module_search_path():
    """Add Piglit's data directory to Python's module search path.

    This enables Python to import Piglit's framework module.

    CAUTION: This script must import the framework that *belongs to this
    script*. Mayhem occurs if this script accidentally imports a framework
    module that belongs to a different Piglit source tree or belongs to
    a different Piglit installation.

    CAUTION: This script file must be located in the Piglit source tree or in
    an installed location.  Otherwise this function may fail to find the
    framework module or, worse, it may succeed in finding a different Piglit's
    framework module.
    """

    piglit_data_dir = os.environ.get('PIGLIT_SOURCE_DIR', None)
    if piglit_data_dir is None:
        print('error: the PIGLIT_SOURCE_DIR env variable is not set. '
              'exiting ...', file=sys.stderr)
        sys.exit(1)

    if path.exists(path.join(piglit_data_dir, 'framework')):
        sys.path.append(piglit_data_dir)
        return

    print('error: failed to find piglit data directory. '
          'exiting...',file=sys.stderr)
    sys.exit(1)


setup_module_search_path()
from framework import status
import framework.replay.programs.compare as compare
import framework.replay.programs.query as query
import framework.replay.programs.download as download
import framework.replay.programs.checksum as checksum
import framework.replay.programs.dump as dump
import framework.replay.programs.profile as profile


def main():
    """ Parse argument and call other executables """
    input_ = sys.argv[1:]

    parser = argparse.ArgumentParser()

    # Add a destination due to
    # https://github.com/python/cpython/pull/3027#issuecomment-330910633
    subparsers = parser.add_subparsers(dest='command', required=True)

    parser_checksum = subparsers.add_parser(
        'checksum',
        add_help=False,
        help=('Calculates the checksums of a local image file.'))
    parser_checksum.set_defaults(func=checksum.checksum)

    parser_compare = subparsers.add_parser(
        'compare',
        add_help=False,
        help=('Trace replay compare methods.'))
    parser_compare.set_defaults(func=compare.compare)

    parser_download = subparsers.add_parser(
        'download',
        add_help=False,
        help=('Downloads a remote file into the db path.'))
    parser_download.set_defaults(func=download.download)

    parser_dump = subparsers.add_parser(
        'dump',
        add_help=False,
        help=('Dumps image(s) from a trace file.'))
    parser_dump.set_defaults(func=dump.dump)

    parser_query = subparsers.add_parser(
        'query',
        add_help=False,
        help=('Queries for specific information '
              'from a traces description file listing traces '
              'and their checksums for each device.'))
    parser_query.set_defaults(func=query.query)

    parser_profile = subparsers.add_parser(
        'profile',
        add_help=False,
        help=('Measure frame times of a trace file.'))
    parser_profile.set_defaults(func=profile.profile)

    # Parse the known arguments (replayer compare or replayer query for
    # example), and then pass the arguments that this parser doesn't know about
    # to that executable
    parsed, args = parser.parse_known_args(input_)
    try:
        runner = parsed.func
    except AttributeError:
        parser.print_help()
        sys.exit(1)

    result = runner(args)
    if result is not status.PASS:
        sys.exit(1 if result is status.CRASH else 2)


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright 2014-2016, 2018-2019 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import abc
import os
import subprocess

from framework import core, grouptools, exceptions
from framework import options
from framework.profile import TestProfile
from framework.test.base import Test, is_crash_returncode, TestRunError

__all__ = [
    'DEQPBaseTest',
    'gen_caselist_txt',
    'iter_deqp_test_cases',
    'make_profile',
]


_EXTRA_ARGS = core.get_option('PIGLIT_DEQP_EXTRA_ARGS',
                              ('deqp', 'extra_args'),
                              default='').split()


def select_source(bin_, filename, mustpass, extra_args):
    """Return either the mustpass list or the generated list."""
    if options.OPTIONS.deqp_mustpass:
        return gen_mustpass_tests(mustpass)
    else:
        return iter_deqp_test_cases(
            gen_caselist_txt(bin_, filename, extra_args))


def make_profile(test_list, test_class):
    """Create a TestProfile instance."""
    profile = TestProfile()
    for testname in test_list:
        # deqp uses '.' as the testgroup separator.
        piglit_name = testname.replace('.', grouptools.SEPARATOR)
        profile.test_list[piglit_name] = test_class(testname)

    return profile


def gen_mustpass_tests(mustpass):
    """Return a testlist from the mustpass list."""
    with open(mustpass, 'r') as f:
        for l in f:
            l = l.strip()
            if l:
                yield l


def gen_caselist_txt(bin_, caselist, extra_args):
    """Generate a caselist.txt and return its path.

    Extra args should be a list of extra arguments to pass to deqp.

    """
    # dEQP is stupid (2014-12-07):
    #   1. To generate the caselist file, dEQP requires that the process's
    #      current directory must be that same as that of the executable.
    #      Otherwise, it fails to find its data files.
    #   2. dEQP creates the caselist file in the process's current directory
    #      and provides no option to change its location.
    #   3. dEQP creates a GL context when generating the caselist. Therefore,
    #      the caselist must be generated on the test target rather than the
    #      build host. In other words, when the build host and test target
    #      differ then we cannot pre-generate the caselist on the build host:
    #      we must *dynamically* generate it during the testrun.
    basedir = os.path.dirname(bin_)
    caselist_path = os.path.join(basedir, caselist)

    # TODO: need to catch some exceptions here...
    with open(os.devnull, 'w') as d:
        env = os.environ.copy()
        env['MESA_GL_VERSION_OVERRIDE'] = '4.6'
        env['MESA_GLES_VERSION_OVERRIDE'] = '3.2'

        subprocess.check_call(
            [bin_, '--deqp-runmode=txt-caselist'] + extra_args, cwd=basedir,
            stdout=d, stderr=d, env=env)
    assert os.path.exists(caselist_path)
    return caselist_path


def iter_deqp_test_cases(case_file):
    """Iterate over original dEQP testcase names."""
    with open(case_file, 'r') as caselist_file:
        for i, line in enumerate(caselist_file):
            if line.startswith('GROUP:'):
                continue
            elif line.startswith('TEST:'):
                yield line[len('TEST:'):].strip()
            else:
                raise exceptions.PiglitFatalError(
                    'deqp: {}:{}: ill-formed line'.format(case_file, i))


class DEQPBaseTest(Test, metaclass=abc.ABCMeta):
    __RESULT_MAP = {
        "Pass": "pass",
        "Fail": "fail",
        "QualityWarning": "warn",
        "InternalError": "fail",
        "Crash": "crash",
        "NotSupported": "skip",
        "ResourceError": "crash",
    }

    @abc.abstractproperty
    def deqp_bin(self):
        """The path to the executable."""

    @abc.abstractproperty
    def extra_args(self):
        """Extra arguments to be passed to the each test instance.

        Needs to return a list, since self.command uses the '+' operator, which
        only works to join two lists together.

        """
        return _EXTRA_ARGS

    def __init__(self, case_name):
        command = [self.deqp_bin, '--deqp-case=' + case_name]

        super(DEQPBaseTest, self).__init__(command)

        # dEQP's working directory must be the same as that of the executable,
        # otherwise it cannot find its data files (2014-12-07).
        # This must be called after super or super will overwrite it
        self.cwd = os.path.dirname(self.deqp_bin)

    @Test.command.getter
    def command(self):
        """Return the command plus any extra arguments."""
        command = super(DEQPBaseTest, self).command
        return command + self.extra_args

    def __find_map(self):
        """Run over the lines and set the result."""
        # splitting this into a separate function allows us to return cleanly,
        # otherwise this requires some break/else/continue madness
        for line in self.result.out.split('\n'):
            line = line.lstrip()
            for k, v in self.__RESULT_MAP.items():
                if line.startswith(k):
                    self.result.result = v
                    return

    def interpret_result(self):
        if is_crash_returncode(self.result.returncode):
            self.result.result = 'crash'
        elif self.result.returncode != 0:
            self.result.result = 'fail'
        else:
            self.__find_map()

        # We failed to parse the test output. Fallback to 'fail'.
        if self.result.result == 'notrun':
            self.result.result = 'fail'

    def _run_command(self, *args, **kwargs):
        """Rerun the command if X11 connection failure happens."""
        for _ in range(5):
            super(DEQPBaseTest, self)._run_command(*args, **kwargs)
            x_err_msg = "FATAL ERROR: Failed to open display"
            if x_err_msg in self.result.err or x_err_msg in self.result.out:
                continue
            return

        raise TestRunError('Failed to connect to X server 5 times', 'fail')



# coding=utf-8
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""Classes dealing with groups of Tests.

In piglit tests are grouped into "profiles", which are equivalent to "suites"
in some other testing nomenclature. A profile is a way to tell the framework
that you have a group of tests you want to run, here are the names of those
tests, and the Test instance.
"""

import ast
import collections
import contextlib
import copy
import gzip
import importlib
import itertools
import multiprocessing
import multiprocessing.dummy
import os
import re
import xml.etree.ElementTree as et

from framework import grouptools, exceptions, status
from framework.dmesg import get_dmesg
from framework.log import LogManager
from framework.monitoring import Monitoring
from framework.test.base import Test, DummyTest
from framework.test.piglit_test import (
    PiglitCLTest, PiglitGLTest, ASMParserTest, BuiltInConstantsTest,
    CLProgramTester, VkRunnerTest, ROOT_DIR,
)
from framework.test.shader_test import ShaderTest, MultiShaderTest
from framework.test.glsl_parser_test import GLSLParserTest
from framework.test.xorg import XTSTest, RendercheckTest
from framework.options import OPTIONS

__all__ = [
    'RegexFilter',
    'TestDict',
    'TestProfile',
    'load_test_profile',
    'run',
]


class RegexFilter(object):
    """An object to be passed to TestProfile.filter.

    This object takes a list (or list-like object) of strings which it converts
    to re.compiled objects (so use raw strings for escape sequences), and acts
    as a callable for filtering tests. If a test matches any of the regex then
    it will be scheduled to run. When the inverse keyword argument is True then
    a test that matches any regex will not be scheduled. Regardless of the
    value of the inverse flag if filters is empty then the test will be run.

    Arguments:
    filters -- a list of regex compiled objects.

    Keyword Arguments:
    inverse -- Inverse the sense of the match.
    """

    def __init__(self, filters, inverse=False):
        self.filters = [re.compile(f, flags=re.IGNORECASE) for f in filters]
        self.inverse = inverse

    def __call__(self, name, _):  # pylint: disable=invalid-name
        # This needs to match the signature (name, test), since it doesn't need
        # the test instance use _.

        # If self.filters is empty then return True, we don't want to remove
        # any tests from the run.
        if not self.filters:
            return True

        if not self.inverse:
            return any(r.search(name) for r in self.filters)
        else:
            return not any(r.search(name) for r in self.filters)


class TestDict(collections.abc.MutableMapping):
    """A special kind of dict for tests.

    This mapping lowers the names of keys by default, and enforces that keys be
    strings (not bytes) and that values are Test derived objects. It is also a
    wrapper around collections.OrderedDict.

    This class doesn't accept keyword arguments, this is intentional. This is
    because the TestDict class is ordered, and keyword arguments are unordered,
    which is a design mismatch.
    """
    def __init__(self):
        # This counter is incremented once when the allow_reassignment context
        # manager is opened, and decremented each time it is closed. This
        # allows stacking of the context manager
        self.__allow_reassignment = 0
        self.__container = collections.OrderedDict()

    def __setitem__(self, key, value):
        """Enforce types on set operations.

        Keys should only be strings, and values should only be Tests.

        This method makes one additional requirement, it lowers the key before
        adding it. This solves a couple of problems, namely that we want to be
        able to use file-system hierarchies as groups in some cases, and those
        are assumed to be all lowercase to avoid problems on case insensitive
        file-systems.
        """
        # keys should be strings
        if not isinstance(key, str):
            raise exceptions.PiglitFatalError(
                "TestDict keys must be strings, but was {}".format(type(key)))

        # Values should either be more Tests
        if not isinstance(value, Test):
            raise exceptions.PiglitFatalError(
                "TestDict values must be a Test, but was a {}".format(
                    type(value)))

        # This must be lowered before the following test, or the test can pass
        # in error if the key has capitals in it.
        key = key.lower()

        # If there is already a test of that value in the tree it is an error
        if not self.__allow_reassignment and key in self.__container:
            if self.__container[key] != value:
                error = (
                    'Further, the two tests are not the same,\n'
                    'The original test has this command:   "{0}"\n'
                    'The new test has this command:        "{1}"'.format(
                        ' '.join(self.__container[key].command),
                        ' '.join(value.command))
                )
            else:
                error = "and both tests are the same."

            raise exceptions.PiglitFatalError(
                "A test has already been assigned the name: {}\n{}".format(
                    key, error))

        self.__container[key] = value

    def __getitem__(self, key):
        """Lower the value before returning."""
        return self.__container[key.lower()]

    def __delitem__(self, key):
        """Lower the value before returning."""
        del self.__container[key.lower()]

    def __len__(self):
        return len(self.__container)

    def __iter__(self):
        return iter(self.__container)

    @contextlib.contextmanager
    def group_manager(self, test_class, group, **default_args):
        """A context manager to make working with flat groups simple.

        This provides a simple way to replace add_plain_test,
        add_concurrent_test, etc. Basic usage would be to use the with
        statement to yield and adder instance, and then add tests.

        This does not provide for a couple of cases.
        1) When you need to alter the test after initialization. If you need to
           set instance.env, for example, you will need to do so manually. It
           is recommended to not use this function for that case, but to
           manually assign the test and set env together, for code clearness.
        2) When you need to use a function that modifies the TestProfile.

        Arguments:
        test_class -- a Test derived class that. Instances of this class will
                      be added to the profile.
        group      -- a string or unicode that will be used as the key for the
                      test in profile.

        Keyword Arguments:
        **         -- any additional keyword arguments will be considered
                      default arguments to all tests added by the adder. They
                      will always be overwritten by **kwargs passed to the
                      adder function

        >>> from framework.test import PiglitGLTest
        >>> p = TestProfile()
        >>> with p.group_manager(PiglitGLTest, 'a') as g:
        ...     g(['test'])
        ...     g(['power', 'test'], 'powertest')
        """
        assert isinstance(group, str), type(group)

        def adder(args, name=None, override_class=None, **kwargs):
            """Helper function that actually adds the tests.

            Arguments:
            args   -- arguments to be passed to the test_class constructor.
                      This must be appropriate for the underlying class

            Keyword Arguments:
            name   -- If this is a a truthy value that value will be used as
                      the key for the test. If name is falsy then args will be
                      ' '.join'd and used as name. Default: None
            kwargs -- Any additional args will be passed directly to the test
                      constructor as keyword args.
            """
            # If there is no name, join the arguments list together to make
            # the name
            if not name:
                assert isinstance(args, list) # //
                name = ' '.join(args)

            assert isinstance(name, str)
            lgroup = grouptools.join(group, name)

            class_ = override_class or test_class

            self[lgroup] = class_(
                args,
                **dict(itertools.chain(default_args.items(), kwargs.items())))

        yield adder

    @property
    @contextlib.contextmanager
    def allow_reassignment(self):
        """Context manager that allows keys to be reassigned.

        Normally reassignment happens in error, but sometimes one actually
        wants to do reassignment, say to add extra options in a reduced
        profile. This method allows reassignment, but only within its context,
        making it an explicit choice to do so.

        It is safe to nest this contextmanager.

        This is not thread safe, or even co-routine safe.
        """
        self.__allow_reassignment += 1
        yield
        self.__allow_reassignment -= 1


class Filters(collections.abc.MutableSequence):

    def __init__(self, iterable=None):
        if iterable:
            self.__container = list(iterable)
        else:
            self.__container = []

    def __getitem__(self, index):
        return self.__container[index]

    def __setitem__(self, index, value):
        self.__container[index] = value

    def __delitem__(self, index):
        del self.__container[index]

    def __len__(self):
        return len(self.__container)

    def __add__(self, other):
        return type(self)(itertools.chain(iter(self), iter(other)))

    def insert(self, index, value):
        self.__container.insert(index, value)

    def run(self, iterable):
        for f in self.__container:
            if hasattr(f, 'reset'):
                f.reset()

        for k, v in iterable:
            if all(f(k, v) for f in self.__container):
                yield k, v


def make_test(element):
    """Rebuild a test instance from xml."""
    def process(elem, opt):
        k = elem.attrib['name']
        v = elem.attrib['value']
        try:
            opt[k] = ast.literal_eval(v)
        except ValueError:
            opt[k] = v

    type_ = element.attrib['type']
    options = {}
    for e in element.findall('./option'):
        process(e, options)
    options['env'] = {e.attrib['name']: e.attrib['value']
                      for e in element.findall('./environment/env')}

    if type_ == 'gl':
        return PiglitGLTest(**options)
    if type_ == 'gl_builtin':
        return BuiltInConstantsTest(**options)
    if type_ == 'cl':
        return PiglitCLTest(**options)
    if type_ == 'cl_prog':
        return CLProgramTester(**options)
    if type_ == 'shader':
        return ShaderTest(**options)
    if type_ == 'glsl_parser':
        return GLSLParserTest(**options)
    if type_ == 'asm_parser':
        return ASMParserTest(**options)
    if type_ == 'vkrunner':
        return VkRunnerTest(**options)
    if type_ == 'multi_shader':
        options['skips'] = []
        for e in element.findall('./Skips/Skip/option'):
            skips = {}
            process(e, skips)
            options['skips'].append(skips)
        return MultiShaderTest(**options)
    if type_ == 'xts':
        return XTSTest(**options)
    if type_ == 'rendercheck':
        return RendercheckTest(**options)
    raise Exception('Unreachable')


class XMLProfile(object):

    def __init__(self, filename):
        self.filename = filename
        self.forced_test_list = []
        self.filters = Filters()
        self.options = {
            'dmesg': get_dmesg(False),
            'monitor': Monitoring(False),
            'ignore_missing': False,
        }

    def __len__(self):
        if not (self.filters or self.forced_test_list):
            with gzip.open(self.filename, 'rt') as f:
                iter_ = et.iterparse(f, events=(b'start', ))
                for _, elem in iter_:
                    if elem.tag == 'PiglitTestList':
                        return int(elem.attrib['count'])
        return sum(1 for _ in self.itertests())

    def setup(self):
        pass

    def teardown(self):
        pass

    def _itertests(self):
        """Always iterates tests instead of using the forced test_list."""
        def _iter():
            with gzip.open(self.filename, 'rt') as f:
                doc = et.iterparse(f, events=(b'end', ))
                _, root = next(doc)  # get the root so we can keep clearing it
                for _, e in doc:
                    if e.tag != 'Test':
                        continue
                    k = e.attrib['name']
                    v = make_test(e)
                    yield k, v
                    root.clear()

        for k, v in self.filters.run(_iter()):
            yield k, v

    def itertests(self):
        if self.forced_test_list:
            alltests = dict(self._itertests())
            opts = collections.OrderedDict()
            for n in self.forced_test_list:
                if self.options['ignore_missing'] and n not in alltests:
                    opts[n] = DummyTest(n, status.NOTRUN)
                else:
                    opts[n] = alltests[n]
            return opts.items()
        else:
            return iter(self._itertests())


class MetaProfile(object):

    """Holds multiple profiles but acts like one.

    This is meant to allow classic profiles like all to exist after being
    split.
    """

    def __init__(self, filename):
        self.forced_test_list = []
        self.filters = Filters()
        self.options = {
            'dmesg': get_dmesg(False),
            'monitor': Monitoring(False),
            'ignore_missing': False,
        }

        tree = et.parse(filename)
        root = tree.getroot()
        self._profiles = [load_test_profile(p.text)
                          for p in root.findall('.//Profile')]

        for p in self._profiles:
            p.options = self.options

    def __len__(self):
        if self.forced_test_list or self.filters:
            return sum(1 for _ in self.itertests())
        return sum(len(p) for p in self._profiles)

    def setup(self):
        pass

    def teardown(self):
        pass

    def _itertests(self):
        def _iter():
            for p in self._profiles:
                for k, v in p.itertests():
                    yield k, v

        for k, v in self.filters.run(_iter()):
            yield k, v

    def itertests(self):
        if self.forced_test_list:
            alltests = dict(self._itertests())
            opts = collections.OrderedDict()
            for n in self.forced_test_list:
                if self.options['ignore_missing'] and n not in alltests:
                    opts[n] = DummyTest(n, status.NOTRUN)
                else:
                    opts[n] = alltests[n]
            return opts.items()
        else:
            return iter(self._itertests())


class TestProfile(object):
    """Class that holds a list of tests for execution.

    This class represents a single testsuite, it has a mapping (dictionary-like
    object) of tests attached (TestDict). This is a mapping of <str>:<Test>
    (python 3 str, python 2 unicode), and the key is delimited by
    grouptools.SEPARATOR.

    The group_manager method provides a context_manager to make adding test to
    the test_list easier, by doing more validation and enforcement.
    >>> t = TestProfile()
    >>> with t.group_manager(Test, 'foo@bar') as g:
    ...     g(['foo'])

    This class does not provide a way to execute itself, instead that is
    handled by the run function in this module, which is able to process and
    run multiple TestProfile objects at once.
    """
    def __init__(self):
        self.test_list = TestDict()
        self.forced_test_list = []
        self.filters = Filters()
        self.options = {
            'dmesg': get_dmesg(False),
            'monitor': Monitoring(False),
            'ignore_missing': False,
        }

    def __len__(self):
        return sum(1 for _ in self.itertests())

    def setup(self):
        """Method to do pre-run setup."""

    def teardown(self):
        """Method to do post-run teardown."""

    def copy(self):
        """Create a copy of the TestProfile.

        This method creates a copy with references to the original instance
        using copy.copy. This allows profiles to be "subclassed" by other
        profiles, without modifying the original.

        copy.deepcopy is used for the filters so the original is
        actually not modified in this case.
        """
        new = copy.copy(self)
        new.test_list = copy.copy(self.test_list)
        new.forced_test_list = copy.copy(self.forced_test_list)
        new.filters = copy.deepcopy(self.filters)
        return new

    def itertests(self):
        """Iterate over tests while filtering.

        This iterator is non-destructive.
        """
        if self.forced_test_list:
            opts = collections.OrderedDict()
            for n in self.forced_test_list:
                if self.options['ignore_missing'] and n not in self.test_list:
                    opts[n] = DummyTest(n, status.NOTRUN)
                else:
                    opts[n] = self.test_list[n]
        else:
            opts = collections.OrderedDict()
            test_keys = list(self.test_list.keys())
            test_keys.sort()
            for k in test_keys:
                opts[k] = self.test_list[k]

        for k, v in self.filters.run(opts.items()):
            yield k, v


def load_test_profile(filename, python=None):
    """Load a python module and return it's profile attribute.

    All of the python test files provide a profile attribute which is a
    TestProfile instance. This loads that module and returns it or raises an
    error.

    This method doesn't care about file extensions as a way to be backwards
    compatible with script wrapping piglit. 'tests/quick', 'tests/quick.tests',
    'tests/quick.py', and 'quick' are all equally valid for filename.

    This will raise a FatalError if the module doesn't exist, or if the module
    doesn't have a profile attribute.

    Raises:
    PiglitFatalError -- if the module cannot be imported for any reason, or if
                        the module lacks a "profile" attribute.

    Arguments:
    filename -- the name of a python module to get a 'profile' from

    Keyword Arguments:
    python -- If this is None (the default) XML is tried, and then a python
              module. If True, then only python is tried, if False then only
              XML is tried.
    """
    name, ext = os.path.splitext(os.path.basename(filename))
    if ext == '.no_isolation':
        name = filename

    if not python:
        # If process-isolation is false then try to load a profile named
        # {name}.no_isolation instead. This is only valid for xml based
        # profiles.
        if ext != '.no_isolation' and not OPTIONS.process_isolation:
            try:
                return load_test_profile(name + '.no_isolation' + ext, python)
            except exceptions.PiglitFatalError:
                # There might not be a .no_isolation version, try to load the
                # regular version in that case.
                pass

        if os.path.isabs(filename):
            if '.meta' in filename:
                return MetaProfile(filename)
            if '.xml' in filename:
                return XMLProfile(filename)

        meta = os.path.join(ROOT_DIR, 'tests', name + '.meta.xml')
        if os.path.exists(meta):
            return MetaProfile(meta)


        xml = os.path.join(ROOT_DIR, 'tests', name + '.xml.gz')
        if os.path.exists(xml):
            return XMLProfile(xml)

    if python is False:
        raise exceptions.PiglitFatalError(
            'Cannot open "tests/{0}.xml.gz" or "tests/{0}.meta.xml"'.format(name))

    try:
        mod = importlib.import_module('tests.{0}'.format(name))
    except ImportError:
        raise exceptions.PiglitFatalError(
            'Failed to import "{}", there is either something wrong with the '
            'module or it doesn\'t exist. Check your spelling?'.format(
                filename))

    try:
        return mod.profile
    except AttributeError:
        raise exceptions.PiglitFatalError(
            'There is no "profile" attribute in module {}.\n'
            'Did you specify the right file?'.format(filename))


def run(profiles, logger, backend, concurrency, jobs):
    """Runs all tests using Thread pool.

    When called this method will flatten out self.tests into self.test_list,
    then will prepare a logger, and begin executing tests through it's Thread
    pools.

    Based on the value of concurrency it will either run all the tests
    concurrently, all serially, or first the thread safe tests then the
    serial tests.

    Finally it will print a final summary of the tests.

    Arguments:
    profiles -- a list of Profile instances.
    logger   -- a log.LogManager instance.
    backend  -- a results.Backend derived instance.
    jobs     -- maximum number of concurrent jobs. Use os.cpu_count() by default
    """
    chunksize = 1

    profiles = [(p, p.itertests()) for p in profiles]
    log = LogManager(logger, sum(len(p) for p, _ in profiles))

    def test(name, test, profile, this_pool=None):
        """Function to call test.execute from map"""
        with backend.write_test(name) as w:
            test.execute(name, log.get(), profile.options)
            w(test.result)
        if profile.options['monitor'].abort_needed:
            this_pool.terminate()

    def run_threads(pool, profile, test_list, filterby=None):
        """ Open a pool, close it, and join it """
        if filterby:
            # Although filterby could be attached to TestProfile as a filter,
            # it would have to be removed when run_threads exits, requiring
            # more code, and adding side-effects
            test_list = (x for x in test_list if filterby(x))

        for n, t in test_list:
            pool.apply_async(test, [n, t, profile, pool])

    def run_profile(profile, test_list):
        """Run an individual profile."""
        profile.setup()
        if concurrency == "all":
            run_threads(multi, profile, test_list)
        elif concurrency == "none":
            run_threads(single, profile, test_list)
        else:
            assert concurrency == "some"
            # test_list is an iterator, we need to copy it to run it twice.
            test_list = itertools.tee(test_list, 2)

            # Filter and return only thread safe tests to the threaded pool
            run_threads(multi, profile, test_list[0],
                        lambda x: x[1].run_concurrent)

            # Filter and return the non thread safe tests to the single
            # pool
            run_threads(single, profile, test_list[1],
                        lambda x: not x[1].run_concurrent)
        profile.teardown()

    # Multiprocessing.dummy is a wrapper around Threading that provides a
    # multiprocessing compatible API
    #
    # The default value of pool is the number of virtual processor cores
    single = multiprocessing.dummy.Pool(1)
    multi = multiprocessing.dummy.Pool(jobs)

    try:
        for p in profiles:
            run_profile(*p)

        for pool in [single, multi]:
            pool.close()
            pool.join()
    finally:
        log.get().summary()

    for p, _ in profiles:
        if p.options['monitor'].abort_needed:
            raise exceptions.PiglitAbort(p.options['monitor'].error_message)



# coding=utf-8
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# This permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR(S) BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
# OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" This module enables the running of GLSL parser tests. """

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import os
import sys
import re
import io

from framework import exceptions, options
from .base import TestIsSkip
from .opengl import FastSkipMixin
from .piglit_test import PiglitBaseTest, TEST_BIN_DIR, ROOT_DIR

__all__ = [
    'GLSLParserTest',
    'GLSLParserNoConfigError',
]

# In different configurations piglit may have one or both of these.
if sys.platform == 'win32':
    _HAS_GL_BIN = os.path.exists(os.path.join(TEST_BIN_DIR, 'glslparsertest.exe'))
    _HAS_GLES_BIN = os.path.exists(os.path.join(TEST_BIN_DIR, 'glslparsertest_gles2.exe'))
else:
    _HAS_GL_BIN = os.path.exists(os.path.join(TEST_BIN_DIR, 'glslparsertest'))
    _HAS_GLES_BIN = os.path.exists(os.path.join(TEST_BIN_DIR, 'glslparsertest_gles2'))

# This forces testing with compatibility extensions, even when GLES support is
# built
_FORCE_DESKTOP_VERSION = os.environ.get('PIGLIT_FORCE_GLSLPARSER_DESKTOP', False)


def _is_gles_version(version):
    """Return True if version is es, otherwise false."""
    assert not isinstance(version, bytes), \
        '{}({})'.format(version, type(version))

    if isinstance(version, str):
        # GLES 3+ versions should have "es" appended, even though
        # glslparsertest doesn't require them. If the version ends in "es" then
        # it is a GLES test for sure.
        if version.endswith('es'):
            return True

        if version.endswith('compatibility'):
            return False

        version = float(version)

    return version in [1.0, 3.0, 3.1, 3.2]


class GLSLParserNoConfigError(exceptions.PiglitInternalError):
    pass


class GLSLParserInternalError(exceptions.PiglitInternalError):
    pass


class Parser(object):
    """Find and parse the config block of a GLSLParserTest.

    Specifically it is necessary to parse a glsl_parser_test to get information
    about it before actually creating a PiglitTest. Even though this could
    be done with a function wrapper, making it a distinct class makes it easier
    to sort in the profile.
    """
    _CONFIG_KEYS = frozenset(['expect_result', 'glsl_version',
                              'require_extensions', 'check_link',
                              'dummy_shader_include', 'shader_include_path'])

    def __init__(self, filepath, installpath=None):
        # a set that stores a list of keys that have been found already
        self.__found_keys = set()
        self.api = None
        self.extensions = set()
        self.shader_version = None
        abs_filepath = os.path.join(ROOT_DIR, filepath)

        try:
            with io.open(abs_filepath, mode='r', encoding='utf-8') as testfile:
                testfile = testfile.read()
                self.config = self.parse(testfile, abs_filepath)
            self.command = self.get_command(filepath, installpath)
        except GLSLParserInternalError as e:
            raise exceptions.PiglitFatalError(
                'In file "{}":\n{}'.format(filepath, str(e)))

        self.set_skip_conditions()

    def set_skip_conditions(self):
        """Set OpenGL and OpenGL ES fast skipping conditions."""
        glsl = self.config['glsl_version']
        if _is_gles_version(glsl):
            self.shader_version = float(glsl[:3])
            if self.shader_version >= 3.0:
                self.api = 'gles3'
            else:
                self.api = 'gles2'
        elif glsl.endswith('compatibility'):
            self.shader_version = float(glsl[:3])
            self.api = 'compat'
        else:
            self.shader_version = float(glsl)
            self.api = 'core'

        req = self.config['require_extensions']
        self.extensions = set(r for r in req.split() if not r.startswith('!'))

        if self.api != 'compat' and 'GL_ARB_compatibility' in self.extensions:
            assert self.api == 'core', 'arb_compat with gles?'
            self.api = 'compat'

        # If GLES is requested, but piglit was not built with a gles version,
        # then ARB_ES3<ver>_compatibility is required. Add it to
        # self.extensions
        if self.api.startswith('gles') and _FORCE_DESKTOP_VERSION:
            if self.shader_version == 1.0:
                ver = '2'
            elif self.shader_version == 3.0:
                ver = '3'
            elif self.shader_version == 3.1:
                ver = '3_1'
            elif self.shader_version == 3.2:
                ver = '3_2'
            ext = 'ARB_ES{}_compatibility'.format(ver)
            self.extensions.add(ext)
            self.command.append(ext)

    @staticmethod
    def pick_binary(version):
        """Pick the correct version of glslparsertest to use.

        This will try to select glslparsertest_gles2 for OpenGL ES tests, and
        glslparsertest for desktop OpenGL tests. However, sometimes this isn't
        possible. In that case all tests will be assigned to the desktop
        version.

        If the test requires desktop OpenGL, but only OpenGL ES is available,
        then the test will be skipped in the python layer.

        """
        if _is_gles_version(version) and not _FORCE_DESKTOP_VERSION:
            return 'glslparsertest_gles2'
        else:
            return 'glslparsertest'

    def get_command(self, filepath, installpath):
        """ Create the command argument to pass to super()

        This private helper creates a configparser object, then reads in the
        provided config (from self.__parser), and tests for required options
        that must be provided. If it does not find them it raises an exception.
        It then crafts a command which is returned, and ultimately passed to
        super()

        """
        for opt in ['expect_result', 'glsl_version']:
            if not self.config.get(opt):
                raise GLSLParserInternalError("Missing required section {} "
                                              "from config".format(opt))

        # Create the command and pass it into a PiglitTest()
        glsl = self.config['glsl_version']
        command = [
            self.pick_binary(glsl),
            installpath or filepath,
            self.config['expect_result'],
            self.config['glsl_version']
        ]

        if self.config['check_link'].lower() == 'true':
            command.append('--check-link')
        if self.config['dummy_shader_include'].lower() == 'true':
            command.append('--dummy-shader-include')
        if self.config['shader_include_path']:
            command.append('--shader-include-path=' + self.config['shader_include_path'])
        command.extend(self.config['require_extensions'].split())

        return command

    def parse(self, testfile, filepath):
        """ Private helper that parses the config file

        This method parses the lines of text file, and then returns a
        StrinIO instance suitable to be parsed by a configparser class.

        It will raise GLSLParserInternalError if any part of the parsing
        fails.

        """
        keys = {'require_extensions': '', 'check_link': 'false', 'dummy_shader_include': 'false',
                'shader_include_path': ''}

        # Text of config section.
        # Create a generator that iterates over the lines in the test file.
        # This allows us to run the loop until we find the header, stop and
        # then run again looking for the config sections.
        # This reduces the need for if statements substantially
        lines = (l.strip() for l in testfile.split('\n'))

        is_header = re.compile(r'(//|/\*|\*)\s*\[config\]')
        for line in lines:
            if is_header.match(line):
                break
        else:
            raise GLSLParserNoConfigError("No [config] section found!")

        is_header = re.compile(r'(//|/\*|\*)\s*\[end config\]')
        is_metadata = re.compile(
            r'(//|/\*|\*)\s*(?P<key>[a-z_]*)\:\s(?P<value>.*)')
        bad_values = re.compile(r'(?![\w\.\! /]).*')

        for line in lines:
            # If strip renendered '' that means we had a blank newline,
            # just go on
            if line in ['', '//', '*']:
                continue
            # If we get to the end of the config break
            elif is_header.match(line):
                break

            match = is_metadata.match(line)
            if match:
                if match.group('key') not in self._CONFIG_KEYS:
                    raise GLSLParserInternalError(
                        "Key {} is not a valid key for a "
                        "glslparser test config block".format(
                            match.group('key')))
                elif match.group('key') in self.__found_keys:
                    # If this key has already been encountered throw an error,
                    # there are no duplicate keys allows
                    raise GLSLParserInternalError(
                        'Duplicate entry for key {}'.format(
                            match.group('key')))
                else:
                    bad = bad_values.search(match.group('value'))
                    # XXX: this always seems to return a match object, even
                    # when the match is ''
                    if bad.group():
                        raise GLSLParserInternalError(
                            'Bad character "{}" at line: "{}". '
                            'Only alphanumerics, _, /, and space '
                            'are allowed'.format(
                                bad.group()[0], line))

                    # Otherwise add the key to the set of found keys, and add
                    # it to the dictionary that will be returned
                    self.__found_keys.add(match.group('key'))
                    keys[match.group('key')] = match.group('value')
            else:
                raise GLSLParserInternalError(
                    "The config section is malformed."
                    "Check file {0} for line {1}".format(filepath, line))
        else:
            raise GLSLParserInternalError("No [end config] section found!")

        return keys


class GLSLParserTest(FastSkipMixin, PiglitBaseTest):
    """A Test derived class specifically for glslparser.

    Arguments:
    filepath -- the path to a glsl_parser_test which must end in .vert,
                .tesc, .tese, .geom or .frag
    """

    def __init__(self, command, api=None, extensions=set(),
                 shader_version=None, env=None, **kwargs):
        super(GLSLParserTest, self).__init__(
            command, run_concurrent=True,
            api=api,
            extensions=extensions,
            shader_version=shader_version,
            env=env)

    @PiglitBaseTest.command.getter
    def command(self):
        command = super(GLSLParserTest, self).command
        
        glslfile = os.path.join(ROOT_DIR, command[1])
        return self.keys() + [command[0], glslfile] + command[2:]

    @classmethod
    def new(cls, filepath, installpath=None):
        """Parse a file and create an instance.

        :param str filepath: the file to parse
        :param Optional[str] installpath:
            The relative path the file will be installed to if different than
            filepath
        """
        parsed = Parser(filepath, installpath)
        return cls(
            parsed.command,
            api=parsed.api,
            extensions=parsed.extensions,
            shader_version=parsed.shader_version)

    def is_skip(self):
        if os.path.basename(self.command[0]) == 'glslparsertest' and not _HAS_GL_BIN:
            raise TestIsSkip('Test is for desktop OpenGL, but piglit was not '
                             'built with OpenGL support.')
        elif (os.path.basename(self.command[0]) == 'glslparsertest_gles2'
              and not _HAS_GLES_BIN):
            raise TestIsSkip('Test is for OpenGL ES, but piglit was not '
                             'built with OpenGL ES support.')

        super(GLSLParserTest, self).is_skip()



# coding=utf-8
#
# Copyright © 2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

import argparse

from framework import exceptions
from framework.replay import image_checksum

__all__ = ['checksum']


def _hexdigest_from_image(args):
    print(image_checksum.hexdigest_from_image(args.file_path))


@exceptions.handler
def checksum(input_):
    """ Parser for replayer checksum command """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'file_path',
        help=('the path to the image file '
              'from which to calculate its checksum'))
    parser.set_defaults(func=_hexdigest_from_image)

    args = parser.parse_args(input_)

    args.func(args)



import os
from urllib.request import url2pathname

import requests


class LocalFileAdapter(requests.adapters.BaseAdapter):
    """Protocol Adapter to allow Requests to GET file:// URLs
    """

    @staticmethod
    def _chkpath(_, path):
        """Return an HTTP status for the given filesystem path."""
        if os.path.isdir(path):
            return 400, "Path Not A File"
        if not os.path.isfile(path):
            return 404, "File Not Found"
        if not os.access(path, os.R_OK):
            return 403, "Access Denied"

        return 200, "OK"

    def send(self, req, **kwargs):  # pylint: disable=unused-argument
        """Return the file specified by the given request

        @type req: C{PreparedRequest}
        """
        path = os.path.normcase(os.path.normpath(url2pathname(req.path_url)))
        response = requests.Response()

        response.status_code, response.reason = self._chkpath(req.method, path)
        if response.status_code == 200 and req.method.lower() != "head":
            try:
                response.raw = open(path, "rb")
                response.raw.release_conn = response.raw.close
            except (OSError, IOError) as err:
                response.status_code = 500
                response.reason = str(err)

        if isinstance(req.url, bytes):
            response.url = req.url.decode("utf-8")
        else:
            response.url = req.url

        response.request = req
        response.connection = self

        return response

    def close(self):
        pass



# coding=utf-8
# Copyright (c) 2016, 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Print each test's command in a user-defined consumable format.

This is meant to provide a mechanism to export piglit commands to other tools,
or to create lists of tests to be consumed via "piglit run --test-list"

"""

import argparse
import os
import sys

from . import parsers
from framework import options, profile, exceptions
from framework.test import Test


def get_command(test, piglit_dir):
    """Get just the name of the command with a path relative to bin."""

    # Make the test command relative to the piglit_dir
    test_command = test.command[:]
    test_command[0] = os.path.relpath(test_command[0], piglit_dir)

    command = ' '.join(test_command)
    return command


@exceptions.handler
def main(input_):
    """The main function."""
    parser = argparse.ArgumentParser(parents=[parsers.CONFIG])
    parser.add_argument("-t", "--include-tests",
                        default=[],
                        action="append",
                        metavar="<regex>",
                        help="Run only matching tests "
                             "(can be used more than once)")
    parser.add_argument("-x", "--exclude-tests",
                        default=[],
                        action="append",
                        metavar="<regex>",
                        help="Exclude matching tests (can be used more than "
                             "once)")
    parser.add_argument("--format",
                        dest="format_string",
                        default="{name} ::: {command}",
                        action="store",
                        help="A template string that defines the output "
                             "format. It has two replacement tokens that can "
                             "be provided, along with any arbitrary text, "
                             "which will be printed verbatim. The two tokens "
                             "are '{name}', which will be replaced with the "
                             "name of the test; and '{command}', which will "
                             "be replaced with the command to run the test.")
    parser.add_argument("testProfile",
                        metavar="<Path to testfile>",
                        help="Path to results folder")
    args = parser.parse_args(input_)

    profile_ = profile.load_test_profile(args.testProfile)

    if args.exclude_tests:
        profile_.filters.append(profile.RegexFilter(args.exclude_tests,
                                                    inverse=True))
    if args.include_tests:
        profile_.filters.append(profile.RegexFilter(args.include_tests))

    # Change to the piglit's path
    piglit_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
    os.chdir(piglit_dir)

    for name, test in profile_.itertests():
        assert isinstance(test, Test)
        print(args.format_string.format(
            name=name,
            command=get_command(test, piglit_dir)))



# coding=utf-8
# Copyright 2014, 2015, 2019 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Piglit integrations for dEQP GLES3 tests."""

import os
import warnings

from framework import core
from framework.test import deqp
from framework.options import OPTIONS

__all__ = ['profile']


def _deprecated_get(env_, conf_, dep_env=None, dep_conf=('', ''), **kwargs):
    """Attempt to get deprecated values, then modern values.

    If a deprecated value is found give the user a warning, this uses
    deqp_get_option internally for both the deprecated and undeprecated paths,
    but prefers the old version and issues a warning if they are encountered.
    The old version is looked up unconditionally, if it is not found then the
    new version will be looked up unconditionally, with the default and
    requires keywords (which the first will not have).
    """
    val = None
    if dep_env is not None and dep_conf is not None:
        val = core.get_option(dep_env, dep_conf)

        if dep_env is not None and os.environ.get(dep_env) is not None:
            # see if the old environment variable was set, if it is uses it,
            # and give a deprecation warning
            warnings.warn(
                '{} has been replaced by {} and will be removed. You should '
                'update any scripts using the old environment variable'.format(
                    dep_env, env_))
        elif dep_conf != ('', '') and core.PIGLIT_CONFIG.has_option(*dep_conf):
            warnings.warn(
                '{} has been replaced by {} and will be removed. You should '
                'update any scripts using the old conf variable'.format(
                    ':'.join(dep_conf), ':'.join(conf_)))

    return val if val is not None else core.get_option(env_, conf_, **kwargs)


_DEQP_GLES3_BIN = _deprecated_get('PIGLIT_DEQP_GLES3_BIN',
                                  ('deqp-gles3', 'bin'),
                                  required=True,
                                  dep_env='PIGLIT_DEQP_GLES3_EXE',
                                  dep_conf=('deqp-gles3', 'exe'))

_DEQP_MUSTPASS = _deprecated_get('PIGLIT_DEQP3_MUSTPASS',
                                 ('deqp-gles3', 'mustpasslist'),
                                 dep_env='PIGLIT_DEQP_MUSTPASS',
                                 required=OPTIONS.deqp_mustpass)

_EXTRA_ARGS = core.get_option('PIGLIT_DEQP_GLES3_EXTRA_ARGS',
                              ('deqp-gles3', 'extra_args'),
                              default='').split()


class DEQPGLES3Test(deqp.DEQPBaseTest):
    deqp_bin = _DEQP_GLES3_BIN

    @property
    def extra_args(self):
        return super(DEQPGLES3Test, self).extra_args + \
            [x for x in _EXTRA_ARGS if not x.startswith('--deqp-case')]


profile = deqp.make_profile(  # pylint: disable=invalid-name
    deqp.select_source(_DEQP_GLES3_BIN, 'dEQP-GLES3-cases.txt', _DEQP_MUSTPASS,
                       _EXTRA_ARGS),
    DEQPGLES3Test)



# coding=utf-8
# Copyright (c) 2015-2016, 2019 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Compression support for backends.

This includes both compression and decompression support.

This provides a low level interface of dictionaries, COMPRESSORS and
DECOMPRESSORS, which use compression modes ('bz2', 'gz', 'xz', 'none') to
provide open-like functions with correct mode settings for writing or reading,
respectively.

They should always take unicode (str in python 3.x) objects. It is up to the
caller to ensure that they're passing unicode and not bytes.

A helper, get_mode(), is provided to return the user selected mode (it will try
the PIGLIT_COMPRESSION environment variable, then the piglit.conf
[core]:compression key, and finally the value of compression.DEFAULT). This is
the best way to get a compressor.

"""

import bz2
import contextlib
import errno
import functools
import gzip
import io
import lzma
import os
import subprocess

from framework import core
from framework import exceptions

__all__ = [
    'UnsupportedCompressor',
    'COMPRESSORS',
    'DECOMPRESSORS',
    'get_mode',
]


class UnsupportedCompressor(exceptions.PiglitInternalError):
    def __init__(self, method, *args, **kwargs):
        super(UnsupportedCompressor, self).__init__(*args, **kwargs)
        self.__method = method

    def __str__(self):
        return 'unsupported compression method {}'.format(self.__method)



DEFAULT = 'bz2'

COMPRESSION_SUFFIXES = ['.gz', '.bz2', '.xz']

COMPRESSORS = {
    'bz2': functools.partial(bz2.open, mode='wt'),
    'gz': functools.partial(gzip.open, mode='wt'),
    'none': functools.partial(open, mode='w'),
    'xz': functools.partial(lzma.open, mode='wt'),
}

DECOMPRESSORS = {
    'bz2': functools.partial(bz2.open, mode='rt'),
    'gz': functools.partial(gzip.open, mode='rt'),
    'none': functools.partial(open, mode='r'),
    'xz': functools.partial(lzma.open, mode='rt'),
}


def get_mode():
    """Return the key value of the correct compressor to use.

    Try the environment variable PIGLIT_COMPRESSION; then check the
    PIGLIT_CONFIG section 'core', option 'compression'; finally fall back to
    DEFAULT.

    This will raise an UnsupportedCompressionError if there isn't a compressor
    for that mode. It is the job of the caller to handle this exceptions

    """
    # This is provided as a function rather than a constant because as a
    # function it can honor changes to the PIGLIT_CONFIG instance, or the
    # PIGLIT_COMPRESSION environment variable.

    method = core.get_option('PIGLIT_COMPRESSION',
                             ('core', 'compression'),
                             default=DEFAULT)

    if method not in COMPRESSORS:
        raise UnsupportedCompressor(method)

    return method



# coding=utf-8
# Copyright (c) 2013-2016, 2019 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" Module implementing classes for reading posix dmesg

Currently this module only has the default DummyDmesg, and a LinuxDmesg.
The method used on Linux requires that timestamps are enabled, and no other
posix system has timestamps.

On OSX and *BSD one would likely want to implement a system that reads the
sysloger, since timestamps can be added by the sysloger, and are not inserted
by dmesg.

Most users do not want to call a specific dmesg implementation themselves,
they will want to use the get_dmesg() helper, which provides the proper
dmesg implementation for their OS.

"""

import abc
import gzip
import re
import subprocess
import sys
import warnings

from framework import exceptions

__all__ = [
    'BaseDmesg',
    'DummyDmesg',
    'LinuxDmesg',
    'get_dmesg',
]


class BaseDmesg(metaclass=abc.ABCMeta):
    """ Abstract base class for Dmesg derived objects

    This provides the bases of the constructor, and most subclasses should call
    super() in __init__(). It provides a concrete implementation of the
    update_result() method, which should be suitible for all subclasses.

    The update_dmesg() method is the primary method that each subclass needs to
    override, as this method is used to to actually read the dmesg ringbuffer,
    and the command used, and the options given are OS dependent.

    This class is not thread safe, because it does not black between the start
    of the test and the reading of dmesg, which means that if two tests run at
    the same time, and test A creates an entri in dmesg, but test B finishes
    first, test B will be marked as having the dmesg error.

    """
    @abc.abstractmethod
    def __init__(self):
        # A list containing all messages since the last time dmesg was read.
        # This is used by update result, and then emptied
        self._new_messages = []

        # If this is set it should be an re.compile() object, which matches
        # entries that are to be considered failures. If an entry does not
        # match the regex it will be ignored. By default (None) all entries are
        # considered
        self.regex = None

        # Populate self.dmesg initially, otherwise the first test will always
        # be full of false positives.
        self.update_dmesg()

    @abc.abstractmethod
    def update_dmesg(self):
        """ Update _new_messages list

        This method should read dmesg, determine which entries are new since
        the last read of dmesg, and update self._new_messages with those
        entries.

        """
        pass

    def update_result(self, result):
        """ Takes a TestResult object and updates it with dmesg statuses

        If dmesg is enabled, and if dmesg has been updated, then replace pass
        with dmesg-warn and warn and fail with dmesg-fail. If dmesg has not
        been updated, it will return the original result passed in.

        Arguments:
        result -- A TestResult instance

        """
        def replace(res):
            """ helper to replace statuses with the new dmesg status

            Replaces pass with dmesg-warn, and warn and fail with dmesg-fail,
            otherwise returns the input

            """
            return {
                "pass": "dmesg-warn",
                "warn": "dmesg-fail",
                "fail": "dmesg-fail"
            }.get(res, res)

        # Get a new snapshot of dmesg
        self.update_dmesg()

        # if update_dmesg() found new entries replace the results of the test
        # and subtests
        if self._new_messages:

            if self.regex:
                for line in self._new_messages:
                    if self.regex.search(line):
                        break
                else:
                    return result

            result.result = replace(result.result)

            # Replace the results of any subtests
            for key, value in result.subtests.items():
                result.subtests[key] = replace(value)

            # Add the dmesg values to the result
            result.dmesg = "\n".join(self._new_messages)

        return result

    def __repr__(self):
        return 'BaseDmesg()'


class LinuxDmesg(BaseDmesg):
    """ Read dmesg on posix systems

    This reads the dmesg ring buffer, stores the contents of the buffer, and
    then compares it whenever called, returning the difference between the
    calls. It requires timestamps to be enabled.

    """
    DMESG_COMMAND = ['dmesg', '--level', 'emerg,alert,crit,err,warn,notice']

    def __init__(self):
        """ Create a dmesg instance """
        # _last_message store the contents of the last entry in dmesg the last
        # time it was read. Because dmesg is a ringbuffer old entries can fall
        # off the end if it becomes too full. To track new entries search for
        # this entry and take any entries that were added after it. This works
        # on Linux because each entry is guaranteed unique by timestamps.
        self._last_message = None
        super(LinuxDmesg, self).__init__()

        # First check to see if we can find the live kernel config.
        try:
            with gzip.open(b"/proc/config.gz", 'r') as f:
                for line in f:
                    if line.startswith(b"CONFIG_PRINTK_TIME=y"):
                        return
        # If there is a malformed or missing file, just ignore it and try the
        # regex match (which may not exist if dmesg ringbuffer was cleared).
        except Exception:
            pass

        if not self._last_message:
            # We need to ensure that there is something in dmesg to check
            # timestamps.
            warnings.warn("Cannot check dmesg for timestamp support. If you "
                          "do not have timestamps enabled in your kernel you "
                          "get incomplete dmesg captures", RuntimeWarning)
        elif not re.match(r'\[\s*\d+\.\d+\]', self._last_message):
            # Do an initial check to ensure that dmesg has timestamps, we need
            # timestamps to work correctly. A proper linux dmesg timestamp
            # looks like this: [    0.00000]
            raise exceptions.PiglitFatalError(
                "Your kernel does not seem to support timestamps, which "
                "are required by the --dmesg option")

    def update_dmesg(self):
        """ Call dmesg using subprocess.check_output

        Get the contents of dmesg, then calculate new messages, finally set
        self.dmesg equal to the just read contents of dmesg.

        """
        dmesg = subprocess.check_output(self.DMESG_COMMAND).decode('utf-8')
        dmesg = dmesg.strip().splitlines()

        # Find all new entries, do this by slicing the list of dmesg to only
        # returns elements after the last element stored. If there are not
        # matches a value error is raised, that means all of dmesg is new
        l = 0
        for index, item in enumerate(reversed(dmesg)):
            if item == self._last_message:
                l = len(dmesg) - index  # don't include the matched element
                break
        self._new_messages = dmesg[l:]

        # Attempt to store the last element of dmesg, unless there was no dmesg
        self._last_message = dmesg[-1] if dmesg else None

    def __repr__(self):
        return 'LinuxDmesg()'


class DummyDmesg(BaseDmesg):
    """ An dummy class for dmesg on non unix-like systems

    This implements a dummy version of the LinuxDmesg, and can be used anytime
    it is not desirable to actually read dmesg, such as non-posix systems, or
    when the contents of dmesg don't matter.

    """
    DMESG_COMMAND = []

    def __init__(self):
        pass

    def update_dmesg(self):
        """ Dummy version of update_dmesg """
        pass

    def update_result(self, result):
        """ Dummy version of update_result """
        return result

    def __repr__(self):
        return 'DummyDmesg()'


def get_dmesg(not_dummy=True):
    """ Return a Dmesg type instance

    Normally this does a system check, and returns the type that proper for
    your system. However, if Dummy is True then it will always return a
    DummyDmesg instance.

    """
    if sys.platform.startswith('linux') and not_dummy:
        return LinuxDmesg()
    return DummyDmesg()



# coding=utf-8
# Copyright (c) 2014-2016, 2019 Intel Corporation
# Copyright © 2020 Valve Corporation.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Module implementing a JUnitBackend for piglit """

import os.path
import re
import shutil
try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree
import json

from framework import grouptools, results, exceptions
from framework.core import PIGLIT_CONFIG
from .abstract import FileBackend
from .register import Registry

__all__ = [
    'REGISTRY',
    'JUnitBackend',
]


_JUNIT_SPECIAL_NAMES = ('api', 'search')

_PID_STR = "pid: "
_START_TIME_STR = "start time: "
_END_TIME_STR = "end time: "

# XML cannot include certain characters. This regex matches the "invalid XML
# text character range".
_FORBIDDEN_XML_TEXT_CHARS_RE = re.compile(
    u'[^\u0009\u000A\u000D\u0020-\uD7FF\uE000-\uFFFD\U00010000-\U0010FFFF]+'
)


def escape_forbidden_xml_text_chars(val, replacement=''):
    """
    Strip invalid XML text characters.
    """
    # See: https://github.com/html5lib/html5lib-python/issues/96
    #
    # The XML 1.0 spec defines the valid character range as:
    # Char ::= #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]
    #
    # Sources:
    # https://www.w3.org/TR/REC-xml/#charsets,
    # https://lsimons.wordpress.com/2011/03/17/stripping-illegal-characters-out-of-xml-in-python/

    return _FORBIDDEN_XML_TEXT_CHARS_RE.sub(replacement, val)


def junit_escape(name):
    name = name.replace('.', '_')
    if name in _JUNIT_SPECIAL_NAMES:
        name += '_'
    return name


class JUnitWriter(object):
    """A class that provides a write mechanism for junit tests."""

    def __init__(self, test_suffix, efail, ecrash):
        self._test_suffix = test_suffix
        self._expected_crashes = ecrash
        self._expected_failures = efail

    @staticmethod
    def _make_names(name):
        """Takes a name from piglit (using grouptools.SEPARATOR and returns a
        split classnam, testname pair in junit format.
        """
        classname, testname = grouptools.splitname(name)
        classname = classname.split(grouptools.SEPARATOR)
        classname = [junit_escape(e) for e in classname]
        classname = '.'.join(classname)

        # Add the test to the piglit group rather than directly to the root
        # group, this allows piglit junit to be used in conjunction with other
        # piglit
        # TODO: It would be nice if other suites integrating with piglit could
        # set different root names.
        classname = 'piglit.' + classname

        return (classname, junit_escape(testname))

    @staticmethod
    def _set_xml_err(element, data, expected_result):
        """Adds the 'system-err' element."""
        err = etree.SubElement(element, 'system-err')
        # We cannot control what is in the error output. Let's escape the
        # forbidden XML characters.
        err.text = escape_forbidden_xml_text_chars(data.err)
        err.text += '\n\n{}{}\n{}{}\n{}{}\n'.format(
            _PID_STR, data.pid,
            _START_TIME_STR, data.time.start,
            _END_TIME_STR, data.time.end)

        if data.result in ['fail', 'dmesg-warn', 'dmesg-fail']:
            if expected_result == "failure":
                err.text += "\n\nWARN: passing test as an expected failure"
            elif expected_result == 'error':
                err.text += \
                    "\n\nERROR: Test should have been crash but was failure"
        elif data.result in ['crash', 'timeout']:
            if expected_result == "error":
                err.text += "\n\nWARN: passing test as an expected crash"
            elif expected_result == 'failure':
                err.text += \
                    "\n\nERROR: Test should have been failure but was crash"
        elif expected_result != "pass":
            err.text += "\n\nERROR: This test passed when it "\
                        "expected {0}".format(expected_result)

    @staticmethod
    def _make_result(element, result, expected_result):
        """Adds the skipped, failure, or error element."""
        res = None
        if result == 'incomplete':
            if expected_result != "error":
                res = etree.SubElement(element, 'failure',
                                       message='Incomplete run.')
            else:
                res = etree.SubElement(element, 'skipped',
                                       message='expected incomplete')
        elif result in ['fail', 'dmesg-warn', 'dmesg-fail']:
            if expected_result == "failure":
                res = etree.SubElement(element, 'skipped',
                                       message='expected failure')
            elif expected_result == 'error':
                res = etree.SubElement(element, 'failure',
                                       message='expected crash, but got '
                                               'failure')
            else:
                res = etree.SubElement(element, 'failure')
        elif result in ['crash', 'timeout']:
            if expected_result == "error":
                res = etree.SubElement(element, 'skipped',
                                       message='expected crash')
            elif expected_result == 'failure':
                res = etree.SubElement(element, 'error',
                                       message='expected failure, but got '
                                               'error')
            else:
                res = etree.SubElement(element, 'error')
        elif expected_result != "pass":
            res = etree.SubElement(element, 'failure',
                                   message="expected {}, but got {}".format(
                                       expected_result, result))
        elif result == 'skip':
            # If the result is skip, then just add the skipped message and go on
            res = etree.SubElement(element, 'skipped')

        # Add the piglit type to the failure result
        if res is not None:
            res.attrib['type'] = str(result)

    def _make_root(self, testname, classname, data):
        """Creates and returns the root element."""
        element = etree.Element('testcase',
                                name=self._make_full_test_name(testname),
                                classname=classname,
                                # Incomplete will not have a time.
                                time=str(data.time.total),
                                status=str(data.result))

        return element

    def _make_full_test_name(self, testname):
        # Jenkins will display special pages when the test has certain names.
        # https://jenkins-ci.org/issue/18062
        # https://jenkins-ci.org/issue/19810
        # The testname variable is used in the calculate_result closure, and
        # must not have the suffix appended.
        return testname + self._test_suffix

    def _expected_result(self, name):
        """Get the expected result of the test."""
        name = name.replace("=", ".").replace(":", ".")
        expected_result = "pass"

        if name in self._expected_failures:
            expected_result = "failure"
            # a test can either fail or crash, but not both
            assert name not in self._expected_crashes

        if name in self._expected_crashes:
            expected_result = "error"

        return expected_result

    def __call__(self, f, name, data):
        classname, testname = self._make_names(name)
        element = self._make_root(testname, classname, data)
        expected_result = self._expected_result(
            '{}.{}'.format(classname, testname).lower())

        # If this is an incomplete status then none of these values will be
        # available, nor
        if data.result != 'incomplete':
            self._set_xml_err(element, data, expected_result)

            # Add stdout
            out = etree.SubElement(element, 'system-out')
            # We cannot control what is in the output. Let's escape the
            # forbidden XML characters.
            out.text = escape_forbidden_xml_text_chars(data.out)

            # Prepend command line to stdout
            out.text = data.command + '\n' + out.text

        self._make_result(element, data.result, expected_result)

        f.write(str(etree.tostring(element).decode('utf-8')))


class JUnitSubtestWriter(JUnitWriter):
    """A JUnitWriter derived class that treats subtest at testsuites.

    This class will turn a piglit test with subtests into a testsuite element
    with each subtest as a testcase element. This subclass is needed because
    not all JUnit readers (like the JUnit plugin for Jenkins) handle nested
    testsuites correctly.
    """

    def _make_root(self, testname, classname, data):
        if data.subtests:
            testname = '{}.{}'.format(classname, testname)
            element = etree.Element('testsuite',
                                    name=testname,
                                    time=str(data.time.total),
                                    tests=str(len(data.subtests)))
            for test, result in data.subtests.items():
                etree.SubElement(element,
                                 'testcase',
                                 name=self._make_full_test_name(test),
                                 classname=testname,
                                 status=str(result))

        else:
            element = super(JUnitSubtestWriter, self)._make_root(
                testname, classname, data)
        return element

    def __call__(self, f, name, data):
        classname, testname = self._make_names(name)
        element = self._make_root(testname, classname, data)

        # If this is an incomplete status then none of these values will be
        # available, nor
        if data.result != 'incomplete':
            self._set_xml_err(element, data, 'pass')

            # Add stdout
            out = etree.SubElement(element, 'system-out')
            # We cannot control what is in the output. Let's escape the
            # forbidden XML characters.
            out.text = escape_forbidden_xml_text_chars(data.out)
            # Prepend command line to stdout
            out.text = data.command + '\n' + out.text

            if data.subtests:
                for subname, result in data.subtests.items():
                    # replace special characters and make case insensitive
                    elem = element.find('.//testcase[@name="{}"]'.format(
                        self._make_full_test_name(subname)))
                    assert elem is not None
                    self._make_result(
                        elem, result,
                        self._expected_result('{}.{}.{}'.format(
                            classname, testname, subname).lower()))
            else:
                self._make_result(element, data.result,
                                  self._expected_result('{}.{}'.format(
                                      classname, testname).lower()))
        else:
            self._make_result(element, data.result,
                              self._expected_result('{}.{}'.format(
                                  classname, testname).lower()))

        f.write(str(etree.tostring(element).decode('utf-8')))


class JUnitBackend(FileBackend):
    """ Backend that produces ANT JUnit XML

    Based on the following schema:
    https://svn.jenkins-ci.org/trunk/hudson/dtkit/dtkit-format/dtkit-junit-model/src/main/resources/com/thalesgroup/dtkit/junit/model/xsd/junit-7.xsd

    """
    _file_extension = 'xml'
    _write = None  # this silences the abstract-not-subclassed warning

    def __init__(self, dest, junit_suffix='', junit_subtests=False, **options):
        super(JUnitBackend, self).__init__(dest, **options)

        # make dictionaries of all test names expected to crash/fail
        # for quick lookup when writing results.  Use lower-case to
        # provide case insensitive matches.
        expected_failures = {}
        if PIGLIT_CONFIG.has_section("expected-failures"):
            for fail, _ in PIGLIT_CONFIG.items("expected-failures"):
                expected_failures[fail.lower()] = True
        expected_crashes = {}
        if PIGLIT_CONFIG.has_section("expected-crashes"):
            for fail, _ in PIGLIT_CONFIG.items("expected-crashes"):
                expected_crashes[fail.lower()] = True

        if not junit_subtests:
            self._write = JUnitWriter(
                junit_suffix, expected_failures, expected_crashes)
        else:
            self._write = JUnitSubtestWriter(  # pylint: disable=redefined-variable-type
                junit_suffix, expected_failures, expected_crashes)

    def initialize(self, metadata):
        """ Do nothing

        Junit doesn't support restore, and doesn't have an initial metadata
        block to write, so all this method does is create the tests directory

        """
        tests = os.path.join(self._dest, 'tests')
        if os.path.exists(tests):
            shutil.rmtree(tests)
        os.mkdir(tests)

    def finalize(self, metadata=None):
        """ Scoop up all of the individual pieces and put them together """
        root = etree.Element('testsuites')
        piglit = etree.Element('testsuite', name='piglit')
        root.append(piglit)
        for each in os.listdir(os.path.join(self._dest, 'tests')):
            with open(os.path.join(self._dest, 'tests', each), 'r') as f:
                # parse returns an element tree, and that's not what we want,
                # we want the first (and only) Element node
                # If the element cannot be properly parsed then consider it a
                # failed transaction and ignore it.
                try:
                    piglit.append(etree.parse(f).getroot())
                except etree.ParseError:
                    continue

        num_tests = len(piglit)
        if not num_tests:
            raise exceptions.PiglitUserError(
                'No tests were run, not writing a result file',
                exitcode=2)

        # set the test count by counting the number of tests.
        # This must be unicode (py3 str)
        piglit.attrib['tests'] = str(num_tests)


        with open(os.path.join(self._dest, 'results.xml'), 'w') as f:
            f.write("<?xml version='1.0' encoding='utf-8'?>\n")
            # lxml has a pretty print we want to use
            if etree.__name__ == 'lxml.etree':
                out = etree.tostring(root, pretty_print=True)
            else:
                out = etree.tostring(root)
            f.write(out.decode('utf-8'))

        shutil.rmtree(os.path.join(self._dest, 'tests'))


def _load(results_file):
    """Load a junit results instance and return a TestrunResult.

    It's worth noting that junit is not as descriptive as piglit's own json
    format, so some data structures will be empty compared to json.

    This tries to not make too many assumptions about the structure of the
    JUnit document.

    """
    run_result = results.TestrunResult()

    splitpath = os.path.splitext(results_file)[0].split(os.path.sep)
    if splitpath[-1] != 'results':
        run_result.name = splitpath[-1]
    elif len(splitpath) > 1:
        run_result.name = splitpath[-2]
    else:
        run_result.name = 'junit result'

    tree = etree.parse(results_file).getroot().find('.//testsuite')
    for test in tree.iterfind('testcase'):
        result = results.TestResult()
        # Take the class name minus the 'piglit.' element, replace junit's '.'
        # separator with piglit's separator, and join the group and test names
        name = test.attrib['name']
        if 'classname' in test.attrib:
            name = grouptools.join(test.attrib['classname'], name)
        name = name.replace('.', grouptools.SEPARATOR)
        is_piglit = False
        if name.startswith("piglit"):
            is_piglit = True
            name = name.split(grouptools.SEPARATOR, 1)[1]

        # Remove the trailing _ if they were added (such as to api and search)
        if name.endswith('_'):
            name = name[:-1]

        result.result = test.attrib['status']

        # This is the fallback path, we'll try to overwrite this with the value
        # in stderr
        result.time = results.TimeAttribute()
        if 'time' in test.attrib:
            result.time = results.TimeAttribute(end=float(test.attrib['time']))
        syserr = test.find('system-err')
        if syserr is not None:
            reversed_err = syserr.text.split('\n')
            reversed_err.reverse()
        else:
            reversed_err = []

        # The command is prepended to system-out, so we need to separate those
        # into two separate elements
        out_tag = test.find('system-out')
        if out_tag is not None:
            if is_piglit:
                out = out_tag.text.split('\n')
                result.command = out[0]
                result.out = '\n'.join(out[1:])
            else:
                result.out = out_tag.text

        err_list = []
        skip = 0
        # Try to get the values in stderr for time and pid
        for line in reversed_err:
            if skip > 0:
                if line == '':
                    skip -= 1
                    continue
                else:
                    skip = 0
            if line.startswith(_START_TIME_STR):
                result.time.start = float(line[len(_START_TIME_STR):])
                continue
            elif line.startswith(_END_TIME_STR):
                result.time.end = float(line[len(_END_TIME_STR):])
                continue
            elif line.startswith(_PID_STR):
                result.pid = json.loads(line[len(_PID_STR):])
                skip = 2
                continue
            err_list.append(line)

        err_list.reverse()
        result.err = "\n".join(err_list)


        run_result.tests[name] = result

    run_result.calculate_group_totals()

    return run_result


def load(results_dir, compression):  # pylint: disable=unused-argument
    """Searches for a results file and returns a TestrunResult.

    wraps _load and searches for the result file.

    """
    if not os.path.isdir(results_dir):
        return _load(results_dir)
    elif os.path.exists(os.path.join(results_dir, 'tests')):
        raise NotImplementedError('resume support of junit not implemented')
    elif os.path.exists(os.path.join(results_dir, 'results.xml')):
        return _load(os.path.join(results_dir, 'results.xml'))
    else:
        raise exceptions.PiglitFatalError("No results found")


def write_results(results, file_, junit_subtests=False):
    """Write the values of the results out to a file."""

    if not junit_subtests:
        writer = JUnitWriter('', {}, {})
    else:
        writer = JUnitSubtestWriter('', {}, {})

    with open(file_, 'w') as f:
        f.write("<?xml version='1.0' encoding='utf-8'?>\n")
        f.write('<testsuites><testsuite name="piglit" tests="{}">'.format(
            len(results.tests)))
        for k, v in results.tests.items():
            writer(f, k, v)
        f.write("</testsuite></testsuites>")

    return False


REGISTRY = Registry(
    extensions=['.xml'],
    backend=JUnitBackend,
    load=load,
    meta=lambda x: x,  # The venerable no-op function
    write=write_results,
)



# coding=utf-8
#
# Copyright (c) 2014-2016, 2019 Intel Corporation
# Copyright © 2019-2020 Valve Corporation.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

"""Provides a module like interface for dump backends.

This package provides an abstract interface for working with dump backends,
which implement various functions as provided in the Registry class, and then
provide a Registry instance as REGISTRY, which maps individual objects to
objects that piglit expects to use. This module then provides a thin
abstraction layer so that piglit is never aware of what backend it's using, it
just asks for an object and receives one.

Most consumers will want to import framework.replay.backends and work directly
with the helper functions here. For some more detailed uses (test cases
especially) the modules themselves may be used.

When this module is loaded it will search through framework/replay/backends for
python modules (those ending in .py), and attempt to import them. Then it will
look for an attribute REGISTRY in those modules, and if it as a
framework.replay.register.Registry instance, it will add the name of that
module (sans .py) as a key, and the instance as a value to the DUMPBACKENDS
dictionary. Each of the helper functions in this module uses that dictionary to
find the function that a user actually wants.

"""

import importlib
import os
from os import path

from .register import Registry

__all__ = [
    'DUMPBACKENDS',
    'DumpBackendError',
    'DumpBackendNotImplementedError',
    'dump',
]


class DumpBackendError(Exception):
    pass


class DumpBackendNotImplementedError(NotImplementedError):
    pass


def _register():
    """Register backends.

    Walk through the list of backends and register them to a name in a
    dictionary, so that they can be referenced from helper functions.

    """
    registry = {}

    for module in os.listdir(path.dirname(path.abspath(__file__))):
        module, extension = path.splitext(module)
        if extension == '.py':
            mod = importlib.import_module(
                'framework.replay.backends.{}'.format(module))
            if hasattr(mod, 'REGISTRY') and isinstance(mod.REGISTRY, Registry):
                registry[module] = mod.REGISTRY

    return registry


DUMPBACKENDS = _register()


def dump(trace_path, output_dir=None, calls=None):
    """Wrapper for dumping traces.

    This function will attempt to determine how to dump the trace file (based
    on file extension), and then pass the trace file path, output_dir and calls
    into the appropriate instance, and then call dump in such instance.

    """
    name, extension = path.splitext(trace_path)

    for dump_backend in DUMPBACKENDS.values():
        if extension in dump_backend.extensions:
            backend = dump_backend.backend

            if backend is None:
                raise DumpBackendNotImplementedError(
                    'DumpBackend for "{}" is not implemented'.format(extension))

            instance = backend(trace_path, output_dir, calls)
            return instance.dump()

    raise DumpBackendError(
        'No module supports file extensions "{}"'.format(extension))
